from __future__ import annotations as _annotations

import asyncio
from dataclasses import dataclass
from pathlib import Path
from types import SimpleNamespace
from typing import Any, cast

import pytest
from acp.exceptions import RequestError
from acp.helpers import text_block
from acp.schema import (
    AllowedOutcome,
    AudioContentBlock,
    BlobResourceContents,
    ContentToolCallContent,
    CreateTerminalResponse,
    EmbeddedResourceContentBlock,
    EnvVariable,
    ImageContentBlock,
    KillTerminalResponse,
    ReadTextFileResponse,
    ReleaseTerminalResponse,
    RequestPermissionResponse,
    ResourceContentBlock,
    SessionConfigOptionBoolean,
    SessionInfoUpdate,
    SessionMode,
    TerminalOutputResponse,
    TextContentBlock,
    TextResourceContents,
    ToolCallStart,
    UserMessageChunk,
    WaitForTerminalExitResponse,
    WriteTextFileResponse,
)
from pydantic import BaseModel
from pydantic_acp.agent_source import AgentFactory, AgentSource
from pydantic_acp.approvals import NativeApprovalBridge
from pydantic_acp.bridges.base import BufferedCapabilityBridge, CapabilityBridge
from pydantic_acp.bridges.history_processor import HistoryProcessorBridge
from pydantic_acp.bridges.prepare_tools import PrepareToolsBridge, PrepareToolsMode
from pydantic_acp.config import AdapterConfig
from pydantic_acp.hook_projection import HookEvent, HookProjectionMap
from pydantic_acp.host.context import ClientHostContext
from pydantic_acp.host.filesystem import ClientFilesystemBackend, FilesystemBackend
from pydantic_acp.host.terminal import ClientTerminalBackend, TerminalBackend
from pydantic_acp.models import AdapterModel
from pydantic_acp.projection import (
    CompositeProjectionMap,
    DefaultToolClassifier,
    FileSystemProjectionMap,
    ProjectionMap,
    ToolClassifier,
    ToolProjection,
    _bash_progress_content,
    _bash_status,
    _candidate_keys,
    _format_command_title,
    _merge_tool_projections,
    _preserve_file_diff_content,
    _preview_text,
    _read_existing_text,
    _single_line_preview,
    _stringify_value,
    build_tool_progress_update,
    build_tool_start_update,
    build_tool_updates,
    compose_projection_maps,
    extract_tool_call_locations,
)
from pydantic_acp.providers import (
    ApprovalStateProvider,
    ConfigOptionsProvider,
    ModelSelectionState,
    ModeState,
    PlanProvider,
    SessionModelsProvider,
    SessionModesProvider,
)
from pydantic_acp.runtime import server as server_module
from pydantic_acp.runtime.bridge_manager import BridgeManager
from pydantic_acp.runtime.prompts import (
    _find_unprocessed_tool_calls,
    _history_contains_text,
    build_cancelled_history,
    build_error_history,
    contains_deferred_tool_requests,
    derive_title,
    dump_message_history,
    load_message_history,
    prompt_to_input,
    prompt_to_text,
    sanitize_message_history,
    usage_from_run,
)
from pydantic_acp.runtime.server import _resolve_config
from pydantic_acp.runtime.session_surface import (
    build_mode_config_option,
    build_mode_state_from_selection,
    build_model_config_option,
    build_model_state_from_selection,
    find_model_option,
)
from pydantic_acp.serialization import DefaultOutputSerializer
from pydantic_acp.session.state import (
    AcpSessionContext,
    StoredSessionUpdate,
    _coerce_json_object,
    _coerce_json_value,
    _is_transcript_kind,
    utc_now,
)
from pydantic_acp.session.store import (
    FileSessionStore,
    MemorySessionStore,
    SessionStore,
    _store_child_path,
)
from pydantic_ai import Agent, ModelRequest, ModelResponse, RunUsage, TextPart
from pydantic_ai.messages import (
    AudioUrl,
    BinaryContent,
    BinaryImage,
    DocumentUrl,
    ImageUrl,
    ModelMessage,
    RetryPromptPart,
    ToolCallPart,
    ToolReturnPart,
    UserPromptPart,
)
from pydantic_ai.models.test import TestModel
from pydantic_ai.tools import DeferredToolRequests, RunContext, ToolApproved, ToolDenied
from typing_extensions import Sentinel

from .support import RecordingClient

_LOW_LEVEL_SENTINEL = Sentinel("_LOW_LEVEL_SENTINEL")


def _session(tmp_path: Path, *, session_id: str = "session-1") -> AcpSessionContext:
    now = utc_now()
    return AcpSessionContext(
        session_id=session_id,
        cwd=tmp_path,
        created_at=now,
        updated_at=now,
    )


def test_default_output_serializer_covers_special_cases() -> None:
    class DemoModel(BaseModel):
        value: int

    @dataclass
    class DemoData:
        value: int

    @dataclass
    class DemoBytesData:
        payload: bytes

    class DemoBytesModel(BaseModel):
        payload: bytes

    class DemoObject:
        def __repr__(self) -> str:
            return "demo-object"

    serializer = DefaultOutputSerializer()

    assert serializer.serialize("plain-text") == "plain-text"
    assert serializer.serialize(b"hi\xff") == "hi\ufffd"
    assert '"value": 1' in serializer.serialize(DemoModel(value=1))
    assert serializer.serialize(DemoData(value=2)) == '{\n  "value": 2\n}'
    assert (
        serializer.serialize(DemoBytesData(payload=b"hi\xff")) == '{\n  "payload": "hi\\ufffd"\n}'
    )
    assert (
        serializer.serialize(DemoBytesModel(payload=b"hi\xff")) == '{\n  "payload": "hi\\ufffd"\n}'
    )
    assert serializer.serialize({"payload": b"hi\xff"}) == '{\n  "payload": "hi\\ufffd"\n}'
    serialized_binary_image = serializer.serialize(
        [BinaryImage(data=b"hi", media_type="image/png")],
    )
    assert "image/png" in serialized_binary_image
    assert (
        serializer.serialize(
            {
                "items": (
                    b"hi\xff",
                    DemoObject(),
                    None,
                    True,
                    3,
                    1.5,
                    "ok",
                ),
            },
        )
        == '{\n  "items": [\n    "hi\\ufffd",\n    "demo-object",\n    null,\n    true,\n    3,\n    1.5,\n    "ok"\n  ]\n}'
    )
    assert serializer.serialize(None) == "null"
    assert serializer.serialize(DemoObject()) == "demo-object"


def test_client_host_context_from_bound_session_requires_client(tmp_path: Path) -> None:
    session = _session(tmp_path)
    with pytest.raises(ValueError):
        ClientHostContext.from_bound_session(session)

    client = RecordingClient()
    session.client = client
    context = ClientHostContext.from_bound_session(session)

    assert context.client is client
    assert context.session is session
    assert context.filesystem.session is session
    assert context.terminal.session is session


def test_protocol_contracts_and_client_backends_cover_interfaces(
    tmp_path: Path,
) -> None:
    session = _session(tmp_path, session_id="protocols")

    class FakeClient:
        def __init__(self) -> None:
            self.calls: list[tuple[str, tuple[Any, ...], dict[str, Any]]] = []

        async def read_text_file(self, *args: Any, **kwargs: Any) -> ReadTextFileResponse:
            self.calls.append(("read_text_file", args, dict(kwargs)))
            return ReadTextFileResponse(content="hello")

        async def write_text_file(self, *args: Any, **kwargs: Any) -> WriteTextFileResponse:
            self.calls.append(("write_text_file", args, dict(kwargs)))
            return WriteTextFileResponse()

        async def create_terminal(self, *args: Any, **kwargs: Any) -> CreateTerminalResponse:
            self.calls.append(("create_terminal", args, dict(kwargs)))
            return CreateTerminalResponse(terminal_id="term-1")

        async def terminal_output(self, *args: Any, **kwargs: Any) -> TerminalOutputResponse:
            self.calls.append(("terminal_output", args, dict(kwargs)))
            return TerminalOutputResponse(output="ok", truncated=False)

        async def release_terminal(self, *args: Any, **kwargs: Any) -> ReleaseTerminalResponse:
            self.calls.append(("release_terminal", args, dict(kwargs)))
            return ReleaseTerminalResponse()

        async def wait_for_terminal_exit(
            self,
            *args: Any,
            **kwargs: Any,
        ) -> WaitForTerminalExitResponse:
            self.calls.append(("wait_for_terminal_exit", args, dict(kwargs)))
            return WaitForTerminalExitResponse(exit_code=0)

        async def kill_terminal(self, *args: Any, **kwargs: Any) -> KillTerminalResponse:
            self.calls.append(("kill_terminal", args, dict(kwargs)))
            return KillTerminalResponse()

    client = FakeClient()
    filesystem = ClientFilesystemBackend(client=cast("Any", client), session=session)
    terminal = ClientTerminalBackend(client=cast("Any", client), session=session)

    assert asyncio.run(filesystem.read_text_file("note.txt", limit=3, line=1)).content == "hello"
    assert asyncio.run(filesystem.write_text_file("note.txt", "body")) is not None
    assert (
        asyncio.run(
            terminal.create_terminal(
                "python",
                args=["-V"],
                cwd=str(tmp_path),
                env=[EnvVariable(name="X", value="1")],
                output_byte_limit=10,
            ),
        ).terminal_id
        == "term-1"
    )
    assert asyncio.run(terminal.terminal_output("term-1")).output == "ok"
    assert asyncio.run(terminal.release_terminal("term-1")) is not None
    assert asyncio.run(terminal.wait_for_terminal_exit("term-1")).exit_code == 0
    assert asyncio.run(terminal.kill_terminal("term-1")) is not None

    assert client.calls[0][0] == "read_text_file"
    assert client.calls[0][2]["session_id"] == "protocols"
    assert client.calls[2][0] == "create_terminal"
    assert client.calls[2][2]["session_id"] == "protocols"

    assert cast("Any", AgentFactory.__call__)(_LOW_LEVEL_SENTINEL, session) is None
    assert asyncio.run(cast("Any", AgentSource.get_agent)(_LOW_LEVEL_SENTINEL, session)) is None
    assert (
        asyncio.run(
            cast("Any", AgentSource.get_deps)(_LOW_LEVEL_SENTINEL, session, _LOW_LEVEL_SENTINEL),
        )
        is None
    )
    assert (
        asyncio.run(cast("Any", FilesystemBackend.read_text_file)(_LOW_LEVEL_SENTINEL, "x")) is None
    )
    assert (
        asyncio.run(cast("Any", FilesystemBackend.write_text_file)(_LOW_LEVEL_SENTINEL, "x", "y"))
        is None
    )
    assert (
        asyncio.run(cast("Any", TerminalBackend.create_terminal)(_LOW_LEVEL_SENTINEL, "bash"))
        is None
    )
    assert (
        asyncio.run(cast("Any", TerminalBackend.terminal_output)(_LOW_LEVEL_SENTINEL, "t")) is None
    )
    assert (
        asyncio.run(cast("Any", TerminalBackend.release_terminal)(_LOW_LEVEL_SENTINEL, "t")) is None
    )
    assert (
        asyncio.run(cast("Any", TerminalBackend.wait_for_terminal_exit)(_LOW_LEVEL_SENTINEL, "t"))
        is None
    )
    assert asyncio.run(cast("Any", TerminalBackend.kill_terminal)(_LOW_LEVEL_SENTINEL, "t")) is None
    assert (
        cast("Any", SessionModelsProvider.get_model_state)(
            _LOW_LEVEL_SENTINEL,
            session,
            _LOW_LEVEL_SENTINEL,
        )
        is None
    )
    assert (
        cast("Any", SessionModelsProvider.set_model)(
            _LOW_LEVEL_SENTINEL,
            session,
            _LOW_LEVEL_SENTINEL,
            "m",
        )
        is None
    )
    assert (
        cast("Any", SessionModesProvider.get_mode_state)(
            _LOW_LEVEL_SENTINEL,
            session,
            _LOW_LEVEL_SENTINEL,
        )
        is None
    )
    assert (
        cast("Any", SessionModesProvider.set_mode)(
            _LOW_LEVEL_SENTINEL,
            session,
            _LOW_LEVEL_SENTINEL,
            "plan",
        )
        is None
    )
    assert (
        cast("Any", ConfigOptionsProvider.get_config_options)(
            _LOW_LEVEL_SENTINEL,
            session,
            _LOW_LEVEL_SENTINEL,
        )
        is None
    )
    assert (
        cast("Any", ConfigOptionsProvider.set_config_option)(
            _LOW_LEVEL_SENTINEL,
            session,
            _LOW_LEVEL_SENTINEL,
            "id",
            True,
        )
        is None
    )
    assert (
        cast("Any", PlanProvider.get_plan)(_LOW_LEVEL_SENTINEL, session, _LOW_LEVEL_SENTINEL)
        is None
    )
    assert (
        cast("Any", ApprovalStateProvider.get_approval_state)(
            _LOW_LEVEL_SENTINEL,
            session,
            _LOW_LEVEL_SENTINEL,
        )
        is None
    )
    assert cast("Any", SessionStore.delete)(_LOW_LEVEL_SENTINEL, "missing") is None
    assert (
        cast("Any", SessionStore.fork)(
            _LOW_LEVEL_SENTINEL,
            "missing",
            new_session_id="copy",
            cwd=tmp_path,
        )
        is None
    )
    assert cast("Any", SessionStore.get)(_LOW_LEVEL_SENTINEL, "missing") is None
    assert cast("Any", SessionStore.list_sessions)(_LOW_LEVEL_SENTINEL) is None
    assert cast("Any", SessionStore.save)(_LOW_LEVEL_SENTINEL, session) is None


def test_file_session_store_covers_delete_fork_missing_and_list_skip(
    tmp_path: Path,
) -> None:
    store = FileSessionStore(tmp_path / "sessions")
    first = _session(tmp_path, session_id="first")
    first.title = "First"
    store.save(first)
    assert store._session_path("first").exists()
    assert store._session_path("safe_1-2").parent == store.root
    assert store._temp_session_path("safe_1-2").parent == store.root
    with pytest.raises(ValueError, match="session path"):
        _store_child_path(store.root, "../escape.json")

    unsafe_session_ids = [
        "",
        "../escape",
        "nested/session",
        ".hidden",
        "two words",
        "x" * 129,
    ]
    for unsafe_session_id in unsafe_session_ids:
        with pytest.raises(ValueError, match="session_id"):
            store._session_path(unsafe_session_id)

    with pytest.raises(ValueError, match="session_id"):
        store.save(_session(tmp_path, session_id="../escape"))
    with pytest.raises(ValueError, match="session_id"):
        store.get("../escape")
    with pytest.raises(ValueError, match="session_id"):
        store.fork("first", new_session_id="../escape", cwd=tmp_path)

    forked = store.fork("first", new_session_id="forked", cwd=tmp_path / "forked")
    assert forked is not None
    assert forked.session_id == "forked"
    assert forked.cwd == tmp_path / "forked"

    store.delete("first")
    assert not store._session_path("first").exists()
    store.delete("first")
    assert store.get("missing") is None
    assert store.fork("missing", new_session_id="nope", cwd=tmp_path) is None

    keep = _session(tmp_path, session_id="keep")
    keep.updated_at = utc_now()
    store.save(keep)

    @dataclass(slots=True)
    class SkipGhostFileSessionStore(FileSessionStore):
        def get(self, session_id: str) -> AcpSessionContext | None:
            if session_id == "ghost":  # pragma: no branch
                return None
            return FileSessionStore.get(self, session_id)

    skip_store = SkipGhostFileSessionStore(store.root)
    ghost_path = skip_store._session_path("ghost")
    ghost_path.write_text("{}", encoding="utf-8")
    (skip_store.root / "bad name.json").write_text("{}", encoding="utf-8")
    assert skip_store.get("ghost") is None
    assert skip_store.get("keep") is not None
    assert [session.session_id for session in skip_store.list_sessions()] == [
        "keep",
        "forked",
    ]

    memory = MemorySessionStore()
    assert memory.get("missing") is None
    assert memory.fork("missing", new_session_id="copy", cwd=tmp_path) is None
    memory.delete("missing")
    store.delete("already-missing")

    stale_path = store.root / ".acpkit-session-stale.tmp"
    stale_path.write_text("stale", encoding="utf-8")
    refreshed_store = FileSessionStore(store.root)
    assert not stale_path.exists()

    with pytest.MonkeyPatch.context() as monkeypatch:
        monkeypatch.setattr(
            "pydantic_acp.session.store.os.open",
            lambda *_args, **_kwargs: (_ for _ in ()).throw(OSError("open failed")),
        )
        refreshed_store._fsync_directory()

    with pytest.MonkeyPatch.context() as monkeypatch:
        monkeypatch.setattr(
            "pydantic_acp.session.store.os.fsync",
            lambda *_args, **_kwargs: (_ for _ in ()).throw(OSError("fsync failed")),
        )
        refreshed_store._fsync_directory()


def test_capability_bridge_defaults_and_buffered_failed_event(tmp_path: Path) -> None:
    session = _session(tmp_path)
    bridge = CapabilityBridge()
    agent = Agent(TestModel())

    assert bridge.drain_updates(session, agent) is None
    assert bridge.get_config_options(session, agent) is None
    assert bridge.get_mcp_capabilities(agent) is None
    assert bridge.get_approval_policy_key("tool") is None
    assert bridge.get_mode_state(session, agent) is None
    assert bridge.get_session_metadata(session, agent) is None
    assert bridge.get_tool_kind("tool") is None
    assert bridge.set_config_option(session, agent, "id", "value") is None
    assert bridge.set_mode(session, agent, "mode") is None

    buffered = BufferedCapabilityBridge()
    buffered._record_failed_event(session, title="failed.event", raw_output="boom")
    updates = buffered.drain_updates(session, agent)
    assert updates is not None
    assert len(updates) == 2


def test_history_processor_bridge_covers_duplicate_names_and_failures(
    tmp_path: Path,
) -> None:
    session = _session(tmp_path)
    bridge = HistoryProcessorBridge()
    messages: list[ModelMessage] = [ModelRequest(parts=[UserPromptPart("hi")])]

    def plain(messages):
        del messages
        raise RuntimeError("plain boom")

    async def contextual(ctx, messages):
        del ctx, messages
        raise RuntimeError("ctx boom")

    wrapped_plain = bridge.wrap_plain_processor(session, plain, name="dup")
    bridge.wrap_plain_processor(session, plain, name="dup")
    wrapped_contextual = bridge.wrap_contextual_processor(session, contextual, name="ctx")

    async def run_plain() -> None:
        await wrapped_plain(messages)

    async def run_contextual() -> None:
        await wrapped_contextual(cast("RunContext[None]", None), messages)

    with pytest.raises(RuntimeError, match="plain boom"):
        asyncio.run(run_plain())
    with pytest.raises(RuntimeError, match="ctx boom"):
        asyncio.run(run_contextual())

    assert bridge.processor_names == ["dup", "ctx"]
    updates = bridge.drain_updates(session, Agent(TestModel()))
    assert updates is not None
    assert len(updates) == 4


def test_prepare_tools_bridge_covers_validation_none_and_require_mode(
    tmp_path: Path,
) -> None:
    with pytest.raises(ValueError, match="at least one mode"):
        PrepareToolsBridge[None](default_mode_id="x", modes=[])
    with pytest.raises(ValueError, match="default mode"):
        PrepareToolsBridge[None](
            default_mode_id="missing",
            modes=[PrepareToolsMode(id="chat", name="Chat", prepare_func=lambda _c, t: t)],
        )

    bridge = PrepareToolsBridge[None](
        default_mode_id="chat",
        modes=[
            PrepareToolsMode(
                id="chat",
                name="Chat",
                prepare_func=cast("Any", lambda _c, _t: None),
            ),
        ],
    )
    session = _session(tmp_path)

    async def run_prepare() -> Any:
        prepared = bridge.build_prepare_tools(session)(cast("RunContext[None]", None), [])
        if asyncio.iscoroutine(prepared):
            return await prepared
        assert prepared is not None  # pragma: no cover
        return prepared  # pragma: no cover

    prepared = asyncio.run(run_prepare())
    assert prepared == []
    assert bridge.current_mode(session).id == "chat"
    assert bridge.is_plan_mode(session) is False
    with pytest.raises(ValueError, match="Unknown prepare-tools mode"):
        bridge._require_mode("missing")
    assert bridge.set_mode(session, Agent(TestModel()), "missing") is None


def test_hook_projection_map_covers_hidden_fallbacks_and_truncation() -> None:
    hidden = HookProjectionMap(hidden_event_ids=frozenset({"hidden"}))
    hidden_event = HookEvent(
        event_id="hidden",
        hook_name="hidden",
        tool_name=None,
        tool_filters=(),
        status="completed",
    )
    assert hidden.build_start_update(tool_call_id="1", event=hidden_event) is None
    assert hidden.build_progress_update(tool_call_id="1", event=hidden_event) is None

    projection = HookProjectionMap(
        event_labels={},
        event_kinds={},
        include_raw_input=False,
        max_output_chars=4,
    )
    event = HookEvent(
        event_id="custom_event",
        hook_name="custom_hook",
        tool_name=None,
        tool_filters=("a",),
        raw_output="abcdef",
        status="completed",
    )
    start = projection.build_start_update(tool_call_id="1", event=event)
    progress = projection.build_progress_update(tool_call_id="1", event=event)
    pending = projection.build_progress_update(
        tool_call_id="1",
        event=HookEvent(
            event_id="custom_event",
            hook_name="custom_hook",
            tool_name=None,
            tool_filters=(),
            status=None,
        ),
    )

    assert start is not None
    assert start.title == "Hook Custom Event (custom_hook)"
    assert start.kind == "execute"
    assert start.raw_input is None
    assert progress is not None
    assert progress.raw_output == "abcd\n\n...[truncated]"
    assert pending is None

    raw_input_projection = HookProjectionMap()
    tool_only_event = HookEvent(
        event_id="tool_event",
        hook_name="",
        tool_name="echo",
        tool_filters=(),
        status="completed",
    )
    raw_input_start = raw_input_projection.build_start_update(
        tool_call_id="2",
        event=tool_only_event,
    )
    assert raw_input_start is not None
    assert raw_input_start.raw_input == {"event": "tool_event", "tool_name": "echo"}

    quiet_projection = HookProjectionMap(
        include_raw_output=False,
        show_hook_name_in_title=False,
        show_tool_name_in_title=False,
    )
    quiet_event = HookEvent(
        event_id="before_tool_execute",
        hook_name="before_tool_execute",
        tool_name="echo",
        tool_filters=(),
        raw_output="done",
        status="completed",
    )
    start_update, progress_update = quiet_projection.build_updates(
        tool_call_id="2",
        event=quiet_event,
    )
    assert start_update is not None
    assert start_update.title == "Hook Before Tool"
    assert progress_update is not None
    assert progress_update.raw_output is None


def test_prompt_history_handles_resources_and_usage_paths() -> None:
    prompt = cast(
        "list[Any]",
        [
            ResourceContentBlock(type="resource_link", name="doc", uri="file:///doc"),
            EmbeddedResourceContentBlock(
                type="resource",
                resource=TextResourceContents(uri="file:///note", text="hello"),
            ),
            ImageContentBlock(type="image", data="aGVsbG8=", mime_type="image/png"),
            AudioContentBlock(type="audio", data="aGVsbG8=", mime_type="audio/wav"),
        ],
    )
    assert derive_title([]) == "Untitled session"
    prompt_text = prompt_to_text(prompt)
    assert "[@doc](file:///doc)" in prompt_text
    assert '<context ref="file:///note">' in prompt_text
    assert "hello" in prompt_text
    assert "[image]" in prompt_to_text(prompt)
    assert "[audio]" in prompt_to_text(prompt)
    assert contains_deferred_tool_requests([str, DeferredToolRequests]) is True
    assert contains_deferred_tool_requests([str, int]) is False
    assert load_message_history(None) == []
    assert usage_from_run(RunUsage()) is None

    unresolved: list[ModelMessage] = [
        ModelResponse(
            parts=[
                TextPart("keep"),
                ToolCallPart("demo_tool", {}, tool_call_id="call-1"),
            ],
        ),
    ]
    sanitized = sanitize_message_history(unresolved, error_text="traceback here")
    assert len(sanitized) == 2
    assert isinstance(sanitized[0], ModelResponse)
    assert isinstance(sanitized[1], ModelResponse)
    assert isinstance(sanitized[1].parts[0], TextPart)
    assert "traceback here" in sanitized[1].parts[0].content

    history = dump_message_history(
        [
            ModelResponse(
                parts=[TextPart("The previous run failed before completion.\n\nTraceback:\nboom")],
            ),
        ],
    )
    error_history = build_error_history(
        history,
        prompt_text="",
        traceback_text="boom",
    )
    loaded = load_message_history(error_history)
    texts = [
        part.content
        for message in loaded
        if isinstance(message, ModelResponse)
        for part in message.parts
        if isinstance(part, TextPart)
    ]
    assert sum("boom" in text for text in texts) == 1

    prompt_input = prompt_to_input(prompt)
    assert isinstance(prompt_input, list)
    assert prompt_input[0] == "[@doc](file:///doc)"
    assert isinstance(prompt_input[1], str)
    assert '<context ref="file:///note">' in prompt_input[1]
    assert isinstance(prompt_input[2], BinaryImage)
    assert prompt_input[2].media_type == "image/png"
    assert isinstance(prompt_input[3], BinaryContent)
    assert prompt_input[3].media_type == "audio/wav"


def test_prompt_to_input_maps_resource_links_and_blob_resources() -> None:
    prompt = cast(
        "list[Any]",
        [
            TextContentBlock(type="text", text="hello"),
            ResourceContentBlock(
                type="resource_link",
                name="image",
                uri="https://example.com/kiwi.png",
                mime_type="image/png",
            ),
            ResourceContentBlock(
                type="resource_link",
                name="audio",
                uri="https://example.com/speech.mp3",
                mime_type="audio/mpeg",
            ),
            EmbeddedResourceContentBlock(
                type="resource",
                resource=BlobResourceContents(
                    uri="resource://doc.pdf",
                    blob="aGVsbG8=",
                    mime_type="application/pdf",
                ),
            ),
        ],
    )

    prompt_input = prompt_to_input(prompt)

    assert isinstance(prompt_input, list)
    assert prompt_input[0] == "hello"
    assert isinstance(prompt_input[1], ImageUrl)
    assert prompt_input[1].url == "https://example.com/kiwi.png"
    assert isinstance(prompt_input[2], AudioUrl)
    assert prompt_input[2].url == "https://example.com/speech.mp3"
    assert isinstance(prompt_input[3], BinaryContent)
    assert prompt_input[3].data == b"hello"
    assert prompt_input[3].media_type == "application/pdf"

    document_prompt = cast(
        "list[Any]",
        [
            ResourceContentBlock(
                type="resource_link",
                name="",
                uri="resource://doc.pdf",
                mime_type="application/pdf",
            ),
            EmbeddedResourceContentBlock(
                type="resource",
                resource=BlobResourceContents(
                    uri="resource://blob.bin",
                    blob="aGVsbG8=",
                    mime_type=None,
                ),
            ),
        ],
    )
    document_input = prompt_to_input(document_prompt)
    assert isinstance(document_input, list)
    assert isinstance(document_input[0], DocumentUrl)
    assert document_input[0].url == "resource://doc.pdf"
    assert isinstance(document_input[1], BinaryContent)
    assert document_input[1].media_type == "application/octet-stream"
    assert (
        prompt_to_text(document_prompt).split("\n\n")[1]
        == "[embedded-resource:resource://blob.bin]"
    )


def test_prompt_history_covers_cancelled_resource_and_blank_text_edges() -> None:
    unknown_resource_prompt = cast(
        "list[Any]",
        [
            ResourceContentBlock(
                type="resource_link",
                name="",
                uri="resource://opaque-item",
                mime_type=None,
            ),
        ],
    )

    assert prompt_to_text(unknown_resource_prompt) == "resource://opaque-item"
    assert prompt_to_input(unknown_resource_prompt) == ["resource://opaque-item"]

    cancelled_history = build_cancelled_history(
        None,
        prompt_text="",
        details_text="cancelled",
    )
    cancelled_messages = load_message_history(cancelled_history)
    cancelled_texts = [
        part.content
        for message in cancelled_messages
        if isinstance(message, ModelResponse)
        for part in message.parts
        if isinstance(part, TextPart)
    ]
    assert cancelled_texts == ["User stopped the run.\n\nRun details:\ncancelled"]

    existing_cancelled = dump_message_history(
        [ModelResponse(parts=[TextPart("User stopped the run.\n\nRun details:\ncancelled")])],
    )
    loaded_cancelled = load_message_history(
        build_cancelled_history(existing_cancelled, prompt_text="", details_text="cancelled"),
    )
    repeated_cancelled_texts = [
        part.content
        for message in loaded_cancelled
        if isinstance(message, ModelResponse)
        for part in message.parts
        if isinstance(part, TextPart)
    ]
    assert repeated_cancelled_texts.count("User stopped the run.\n\nRun details:\ncancelled") == 1
    assert usage_from_run(RunUsage(details={"reasoning_tokens": 5})) is not None
    assert _history_contains_text([], "   ") is False

    tool_call = ToolCallPart(tool_name="demo", args={}, tool_call_id="call-1")
    tool_result = ToolReturnPart(tool_name="demo", tool_call_id="call-1", content="done")
    assert _find_unprocessed_tool_calls([ModelResponse(parts=[tool_call])]) == [tool_call]
    assert (
        _find_unprocessed_tool_calls(
            [ModelResponse(parts=[tool_call]), ModelRequest(parts=[tool_result])],
        )
        == []
    )
    assert (
        _find_unprocessed_tool_calls(
            [ModelRequest(parts=[tool_result]), ModelResponse(parts=[tool_call])],
        )
        == []
    )
    assert _find_unprocessed_tool_calls(
        [cast("Any", SimpleNamespace()), ModelResponse(parts=[tool_call])],
    ) == [tool_call]


def test_prompt_to_input_keeps_text_only_prompt_as_string() -> None:
    prompt = [text_block("first"), text_block("second")]

    prompt_input = prompt_to_input(prompt)

    assert prompt_input == "first\n\nsecond"


def test_prompt_to_input_preserves_zed_git_diff_selection_context() -> None:
    uri = "zed:///agent/git-diff?base=main"
    prompt = cast(
        "list[Any]",
        [
            EmbeddedResourceContentBlock(
                type="resource",
                resource=TextResourceContents(
                    uri=uri,
                    text="diff --git a/app.py b/app.py\n@@ -1 +1 @@\n-old\n+new",
                    mime_type="text/x-diff",
                ),
            ),
        ],
    )

    prompt_input = prompt_to_input(prompt)
    prompt_text = prompt_to_text(prompt)

    assert isinstance(prompt_input, list)
    assert prompt_input == [prompt_text]
    assert "[@git-diff?base=main](zed:///agent/git-diff?base=main)" in prompt_text
    assert '<context ref="zed:///agent/git-diff?base=main">' in prompt_text
    assert "diff --git a/app.py b/app.py" in prompt_text


def test_prompt_to_input_preserves_file_refs_rule_text_and_directory_links() -> None:
    hook_uri = (
        "file:///workspace/acpkit/packages/adapters/pydantic-acp/src/"
        "pydantic_acp/bridges/_hook_capability.py?symbol=wrap_run#L79:79"
    )
    thread_executor_uri = (
        "file:///workspace/acpkit/references/pydantic-ai-latest/"
        "pydantic_ai_slim/pydantic_ai/capabilities/thread_executor.py#L54:54"
    )
    coverage_uri = "file:///workspace/acpkit/COVERAGE"
    repo_uri = "file:///workspace/acpkit"
    prompt = cast(
        "list[Any]",
        [
            TextContentBlock(type="text", text="@rule iyi kod yaz"),
            EmbeddedResourceContentBlock(
                type="resource",
                resource=TextResourceContents(
                    uri=hook_uri,
                    text="    async def wrap_run(\n",
                    mime_type="text/x-python",
                ),
            ),
            EmbeddedResourceContentBlock(
                type="resource",
                resource=TextResourceContents(
                    uri=thread_executor_uri,
                    text="executor\n",
                    mime_type="text/plain",
                ),
            ),
            EmbeddedResourceContentBlock(
                type="resource",
                resource=TextResourceContents(
                    uri=coverage_uri,
                    text=(
                        "Line coverage: 97.10% (3930 / 4021)\n"
                        "Branch coverage: 95.03% (1186 / 1248)\n"
                    ),
                    mime_type="text/plain",
                ),
            ),
            ResourceContentBlock(
                type="resource_link",
                name="acpkit",
                uri=repo_uri,
            ),
        ],
    )

    prompt_input = prompt_to_input(prompt)

    assert isinstance(prompt_input, list)
    assert prompt_input[0] == "@rule iyi kod yaz"
    hook_context = prompt_input[1]
    thread_context = prompt_input[2]
    coverage_context = prompt_input[3]
    repo_link = prompt_input[4]
    assert isinstance(hook_context, str)
    assert isinstance(thread_context, str)
    assert isinstance(coverage_context, str)
    assert isinstance(repo_link, str)
    assert f"[@_hook_capability.py?symbol=wrap_run#L79:79]({hook_uri})" in hook_context
    assert f'<context ref="{hook_uri}">' in hook_context
    assert "    async def wrap_run(" in hook_context
    assert hook_context.endswith("</context>")
    assert f"[@thread_executor.py#L54:54]({thread_executor_uri})" in thread_context
    assert f'<context ref="{thread_executor_uri}">' in thread_context
    assert "executor" in thread_context
    assert thread_context.endswith("</context>")
    assert f"[@COVERAGE]({coverage_uri})" in coverage_context
    assert f'<context ref="{coverage_uri}">' in coverage_context
    assert "Line coverage: 97.10% (3930 / 4021)" in coverage_context
    assert "Branch coverage: 95.03% (1186 / 1248)" in coverage_context
    assert coverage_context.endswith("</context>")
    assert repo_link == "[@acpkit](file:///workspace/acpkit)"


def test_session_state_round_trips_and_rejects_invalid_payloads(tmp_path: Path) -> None:
    assert _is_transcript_kind("tool_call")
    assert not _is_transcript_kind("nope")

    update = StoredSessionUpdate.from_update(
        UserMessageChunk(
            session_update="user_message_chunk",
            content=text_block("hello"),
            message_id="msg-1",
        ),
    )
    restored = update.to_update()
    assert restored.session_update == "user_message_chunk"
    session_info_update = StoredSessionUpdate.from_update(
        SessionInfoUpdate(
            session_update="session_info_update",
            title="Demo",
            updated_at=utc_now().isoformat(),
            field_meta={"demo": {"ok": True}},
        ),
    )
    assert session_info_update.to_update().session_update == "session_info_update"

    class BadUpdate:
        def model_dump(self, **kwargs):
            del kwargs
            return {"nope": "missing"}

    with pytest.raises(TypeError, match="missing `sessionUpdate`"):
        StoredSessionUpdate.from_update(cast("Any", BadUpdate()))
    with pytest.raises(TypeError, match="Expected a JSON object"):
        _coerce_json_object("x")
    with pytest.raises(TypeError, match="keys must be strings"):
        _coerce_json_object({1: "x"})
    with pytest.raises(TypeError, match="Unsupported JSON value"):
        _coerce_json_value(_LOW_LEVEL_SENTINEL)
    with pytest.raises(AssertionError):
        StoredSessionUpdate(kind=cast("Any", "unknown"), payload={}).to_update()


def test_session_surface_helpers_handle_missing_state_and_errors() -> None:
    model = AdapterModel(model_id="demo", name="Demo", override="demo")
    with pytest.raises(RequestError):
        build_model_config_option(
            ModelSelectionState(
                current_model_id=None,
                config_option_name="Model",
                config_option_description="Select",
                available_models=[model],
            ),
        )
    with pytest.raises(RequestError):
        build_mode_config_option(ModeState(current_mode_id=None, modes=[]))

    assert build_model_state_from_selection(None) is None
    assert build_mode_state_from_selection(None) is None
    assert build_mode_state_from_selection(ModeState(current_mode_id=None, modes=[])) is None
    assert find_model_option("missing", available_models=[model]) is None
    mode_state = build_mode_state_from_selection(
        ModeState(
            current_mode_id="plan",
            modes=[SessionMode(id="plan", name="Plan", description="d")],
        ),
    )
    assert mode_state is not None
    assert mode_state.current_mode_id == "plan"


def test_server_helpers_resolve_config_and_dispatch_run_acp(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    projection = FileSystemProjectionMap(default_read_tool="read_file")
    hook_projection = HookProjectionMap()

    resolved = _resolve_config(
        config=AdapterConfig(agent_name="custom"),
        agent_name="agent-name",
        projection_maps=[projection, hook_projection],
    )
    assert resolved.agent_name == "custom"
    assert resolved.hook_projection_map is hook_projection
    assert resolved.projection_maps == (projection,)

    resolved_default = _resolve_config(
        config=None,
        agent_name=None,
        projection_maps=None,
    )
    assert resolved_default.agent_name == "pydantic-acp"

    seen: dict[str, Any] = {}

    def fake_create_acp_agent(*args, **kwargs):
        del args, kwargs
        return "adapter"

    async def fake_run_agent(adapter):
        seen["adapter"] = adapter

    def fake_asyncio_run(coro):
        asyncio.run(coro)

    monkeypatch.setattr(server_module, "create_acp_agent", fake_create_acp_agent)
    monkeypatch.setattr(server_module, "run_agent", fake_run_agent)
    monkeypatch.setattr(server_module, "asyncio", SimpleNamespace(run=fake_asyncio_run))

    server_module.run_acp(agent=Agent(TestModel()))
    assert seen["adapter"] == "adapter"


def test_bridge_manager_merges_metadata_and_classification_fallbacks(
    tmp_path: Path,
) -> None:
    session = _session(tmp_path)
    agent = Agent(TestModel(custom_output_text="unused"))

    class MetadataOnlyBridge(CapabilityBridge):
        metadata_key = None

        def get_session_metadata(self, session, agent):
            del session, agent  # pragma: no cover
            return cast("Any", {"hidden": True})  # pragma: no cover

    class FullBridge(CapabilityBridge):
        metadata_key = "visible"

        def drain_updates(self, session, agent):
            del session, agent
            return cast("Any", [_LOW_LEVEL_SENTINEL])

        def get_config_options(self, session, agent):
            del session, agent
            return cast(
                "Any",
                [
                    SessionConfigOptionBoolean(
                        id="flag",
                        name="Flag",
                        type="boolean",
                        current_value=True,
                    ),
                ],
            )

        def get_mcp_capabilities(self, agent=None):
            del agent
            return cast("Any", SimpleNamespace(http=True, sse=False))

        def get_mode_state(self, session, agent):
            del session, agent
            return ModeState(current_mode_id="plan", modes=[SessionMode(id="plan", name="Plan")])

        def get_session_metadata(self, session, agent):
            del session, agent
            return cast("Any", {"ok": True})

        def set_config_option(self, session, agent, config_id, value):
            del session, agent
            if config_id != "flag":
                return None
            return cast(
                "Any",
                [
                    SessionConfigOptionBoolean(
                        id="flag",
                        name="Flag",
                        type="boolean",
                        current_value=bool(value),
                    ),
                ],
            )

        def set_mode(self, session, agent, mode_id):
            del session, agent
            if mode_id != "plan":
                return None
            return ModeState(current_mode_id="plan", modes=[SessionMode(id="plan", name="Plan")])

        def get_tool_kind(self, tool_name, raw_input=None):
            del raw_input
            return "execute" if tool_name == "bridge-tool" else None

        def get_approval_policy_key(self, tool_name, raw_input=None):
            del raw_input
            return "bridge:key" if tool_name == "bridge-tool" else None

    class EmptyMetadataBridge(CapabilityBridge):
        metadata_key = "ignored"

        def get_session_metadata(self, session, agent):
            del session, agent

    manager = BridgeManager(
        base_classifier=DefaultToolClassifier(),
        bridges=[MetadataOnlyBridge(), FullBridge(), EmptyMetadataBridge()],
    )

    assert manager.drain_updates(session, agent) is not None
    assert manager.get_config_options(session, agent) is not None
    caps = manager.get_mcp_capabilities(agent)
    assert caps.http is True
    assert caps.sse is False
    assert manager.get_mode_state(session, agent) is not None
    assert manager.get_metadata_sections(session, agent) == {"visible": {"ok": True}}
    assert manager.set_config_option(session, agent, "missing", True) is None
    assert manager.set_config_option(session, agent, "flag", False) is not None
    assert manager.set_mode(session, agent, "missing") is None
    assert manager.set_mode(session, agent, "plan") is not None
    assert manager.tool_classifier.classify("bridge-tool") == "execute"
    assert manager.tool_classifier.classify("other-tool") == "execute"
    assert manager.tool_classifier.approval_policy_key("bridge-tool") == "bridge:key"
    assert manager.tool_classifier.approval_policy_key("other-tool") == "other-tool"


def test_projection_helpers_handle_fallback_and_edge_paths(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    projection = FileSystemProjectionMap(
        default_write_tool="write_file",
        default_read_tool="read_file",
        default_bash_tool="run_command",
    )
    assert compose_projection_maps(None) is None
    assert compose_projection_maps([]) is None
    assert compose_projection_maps([projection]) is projection
    assert isinstance(compose_projection_maps([projection, projection]), CompositeProjectionMap)

    assert projection.project_start("run_command", raw_input=[]) is None
    assert projection.project_start("run_command", raw_input={"cmd": ""}) is None
    assert projection.project_start("write_file", raw_input=[]) is None
    assert projection.project_start("write_file", raw_input={"path": ""}) is None
    assert (
        projection.project_progress(
            "write_file",
            raw_input={"path": "x", "content": "y"},
            raw_output=None,
            serialized_output="",
            status="failed",
        )
        is None
    )
    assert (
        projection.project_progress(
            "read_file",
            raw_input=[],
            raw_output=None,
            serialized_output="",
            status="completed",
        )
        is None
    )
    assert (
        projection.project_progress(
            "read_file",
            raw_input={"path": ""},
            raw_output=None,
            serialized_output="",
            status="completed",
        )
        is None
    )

    assert extract_tool_call_locations([]) is None
    assert extract_tool_call_locations({"other": "x"}) is None
    assert _candidate_keys(None, ("a", "b")) == ("a", "b")
    assert _candidate_keys("x", ("a", "b")) == ("x", "a", "b")
    assert cast("Any", ToolClassifier.classify)(_LOW_LEVEL_SENTINEL, "tool") is None
    assert cast("Any", ToolClassifier.approval_policy_key)(_LOW_LEVEL_SENTINEL, "tool") is None
    assert cast("Any", ProjectionMap.project_start)(_LOW_LEVEL_SENTINEL, "tool") is None
    assert (
        cast("Any", ProjectionMap.project_progress)(
            _LOW_LEVEL_SENTINEL,
            "tool",
            serialized_output="x",
            status="completed",
        )
        is None
    )
    assert _merge_tool_projections([None]) is None
    merged = _merge_tool_projections(
        [
            ToolProjection(title="first"),
            ToolProjection(status="failed"),
        ],
    )
    assert merged is not None
    assert merged.title == "first"
    assert merged.status == "failed"

    known_start = build_tool_start_update(
        ToolCallPart("write_file", {"path": "a", "content": "b"}, tool_call_id="1"),
        classifier=DefaultToolClassifier(),
        projection_map=projection,
        cwd=tmp_path,
    )
    keep_diff = _preserve_file_diff_content(
        known_start=known_start,
        projection=ToolProjection(content=known_start.content),
    )
    assert keep_diff == known_start.content
    assert (
        _preserve_file_diff_content(known_start=None, projection=ToolProjection(content=[])) == []
    )
    assert _preserve_file_diff_content(
        known_start=known_start,
        projection=ToolProjection(
            content=[ContentToolCallContent(type="content", content=text_block("x"))],
        ),
    ) == [ContentToolCallContent(type="content", content=text_block("x"))]
    assert (
        _preserve_file_diff_content(
            known_start=known_start,
            projection=ToolProjection(content=[]),
        )
        == []
    )
    assert (
        _preserve_file_diff_content(
            known_start=ToolCallStart(
                session_update="tool_call",
                tool_call_id="1",
                title="write_file",
                kind="edit",
                status="in_progress",
                content=[ContentToolCallContent(type="content", content=text_block("x"))],
            ),
            projection=ToolProjection(content=known_start.content),
        )
        == known_start.content
    )
    assert projection._terminal_id_from_value("bad") is None
    assert projection._terminal_id_from_value({"terminalId": "abc"}) == "abc"

    assert _stringify_value("x", None) == "x"
    assert _stringify_value(1, "fallback") == "fallback"
    assert _stringify_value(1, None) == "1"
    assert _read_existing_text("relative.txt", cwd=None) == ""
    assert _read_existing_text("missing.txt", cwd=tmp_path) == ""
    bad_file = tmp_path / "bad.txt"
    bad_file.write_text("content", encoding="utf-8")
    monkeypatch.setattr(Path, "read_text", lambda *args, **kwargs: (_ for _ in ()).throw(OSError()))
    assert _read_existing_text(str(bad_file), cwd=tmp_path) == ""

    assert _preview_text("x" * 5000).endswith("...[truncated]")
    assert _single_line_preview("a\nb", limit=10) == "a b"
    assert _single_line_preview("x" * 20, limit=5) == "xxxxx..."
    assert _format_command_title("echo hi").startswith("Execute ")
    assert _bash_status(None, fallback="completed") == "completed"
    assert _bash_status({"timed_out": True}, fallback="completed") == "failed"
    assert _bash_status({"timed_out": 1}, fallback="completed") == "failed"
    assert _bash_status({"returncode": 1}, fallback="completed") == "failed"
    assert DefaultToolClassifier().classify("delete_file") == "delete"
    assert DefaultToolClassifier().classify("move_file") == "move"
    assert DefaultToolClassifier().classify("search_repo") == "search"
    assert DefaultToolClassifier().classify("fetch_url") == "fetch"
    assert DefaultToolClassifier().classify("think_step") == "think"

    content = _bash_progress_content(
        raw_input={"command": "echo hi"},
        raw_output={"returncode": 2, "stdout": "out", "stderr": "err"},
        serialized_output="serialized",
    )
    assert "Status: failed" in content[0].content.text
    assert "Stdout:" in content[0].content.text
    assert "Stderr:" in content[0].content.text
    fallback_content = _bash_progress_content(
        raw_input=None,
        raw_output=None,
        serialized_output="serialized",
    )
    assert fallback_content[0].content.text == "serialized"
    success_content = _bash_progress_content(
        raw_input={"other": "x"},
        raw_output={"returncode": 0},
        serialized_output="serialized",
    )
    assert "Status: success" in success_content[0].content.text

    retry = RetryPromptPart(
        tool_name="write_file",
        tool_call_id="1",
        content="try again",
    )
    progress = build_tool_progress_update(
        retry,
        classifier=DefaultToolClassifier(),
        known_start=known_start,
        projection_map=projection,
        serializer=DefaultOutputSerializer(),
    )
    assert progress.status == "failed"

    duplicate_updates = build_tool_updates(
        [
            ModelResponse(
                parts=[ToolCallPart("write_file", {"path": "a", "content": "b"}, tool_call_id="1")],
            ),
            ModelRequest(
                parts=[
                    ToolReturnPart("final_result", "done"),
                    RetryPromptPart("retry", tool_name=None),
                ],
            ),
        ],
        classifier=DefaultToolClassifier(),
        known_starts={"1": known_start},
        projection_map=projection,
        serializer=DefaultOutputSerializer(),
    )
    assert duplicate_updates == []


def test_native_approval_bridge_handles_remembered_policies_and_invalid_options(
    tmp_path: Path,
) -> None:
    session = _session(tmp_path)
    tool_call = ToolCallPart("write_file", {"path": "demo.txt", "content": "x"}, tool_call_id="1")
    requests = DeferredToolRequests(calls=[], approvals=[tool_call], metadata={})
    classifier = DefaultToolClassifier()
    bridge = NativeApprovalBridge(enable_persistent_choices=True)

    session.metadata["approval_policies"] = {"write_file": "allow"}
    resolution = asyncio.run(
        bridge.resolve_deferred_approvals(
            client=RecordingClient(),
            session=session,
            requests=requests,
            classifier=classifier,
        ),
    )
    assert isinstance(resolution.deferred_tool_results.approvals["1"], ToolApproved)
    session.metadata["approval_policies"] = {"write_file": "reject"}
    rejected = asyncio.run(
        bridge.resolve_deferred_approvals(
            client=RecordingClient(),
            session=session,
            requests=requests,
            classifier=classifier,
        ),
    )
    assert isinstance(rejected.deferred_tool_results.approvals["1"], ToolDenied)

    class InvalidOptionClient(RecordingClient):
        async def request_permission(self, *args, **kwargs):
            del args, kwargs
            return RequestPermissionResponse(
                outcome=AllowedOutcome(outcome="selected", option_id="bad-option"),
            )

    session.metadata["approval_policies"] = "bad"
    with pytest.raises(RequestError):
        asyncio.run(
            bridge.resolve_deferred_approvals(
                client=InvalidOptionClient(),
                session=session,
                requests=requests,
                classifier=classifier,
            ),
        )

    disabled = NativeApprovalBridge(enable_persistent_choices=False)
    disabled._remember_policy(
        session=session,
        approval_policy_key="write_file",
        option_id="allow_always",
    )
    assert disabled._approval_policies(session) == {}
    assert bridge._selected_option_to_result("unknown") is None
    assert isinstance(bridge._policy_to_result("allow"), ToolApproved)
    assert isinstance(bridge._policy_to_result("reject"), ToolDenied)
    bridge._remember_policy(
        session=session,
        approval_policy_key="write_file",
        option_id="reject_always",
    )
    assert bridge._get_remembered_policy(session, "write_file") == "reject"
    session.metadata["approval_policies"] = {}
    bridge._remember_policy(
        session=session,
        approval_policy_key="write_file",
        option_id="allow_always",
    )
    assert bridge._get_remembered_policy(session, "write_file") == "allow"
    bridge._remember_policy(
        session=session,
        approval_policy_key="write_file",
        option_id="ignore_once",
    )
    bridge._set_remembered_policy(
        session=session,
        approval_policy_key="write_file",
        policy="reject",
    )
    assert bridge._approval_policies(session)["write_file"] == "reject"


def test_build_tool_progress_update_uses_projected_title_without_known_start() -> None:
    @dataclass(slots=True)
    class TitleOnlyProjectionMap:
        title: str

        def project_start(
            self,
            tool_name: str,
            *,
            cwd: Path | None = None,
            raw_input: Any = None,
        ) -> ToolProjection | None:
            del tool_name, cwd, raw_input  # pragma: no cover
            return None  # pragma: no cover

        def project_progress(
            self,
            tool_name: str,
            *,
            cwd: Path | None = None,
            raw_input: Any = None,
            raw_output: Any = None,
            serialized_output: str,
            status: str,
        ) -> ToolProjection | None:
            del tool_name, cwd, raw_input, raw_output, serialized_output, status
            return ToolProjection(title=self.title)

    progress = build_tool_progress_update(
        ToolReturnPart("read_file", "payload", tool_call_id="read-1"),
        classifier=DefaultToolClassifier(),
        known_start=None,
        projection_map=TitleOnlyProjectionMap(title="Projected Read"),
        serializer=DefaultOutputSerializer(),
    )

    assert progress.title == "Projected Read"
    assert progress.locations is None
    assert progress.raw_output == "payload"


def test_build_tool_updates_filters_output_parts_and_empty_messages() -> None:
    classifier = DefaultToolClassifier()
    serializer = DefaultOutputSerializer()

    assert (
        build_tool_updates(
            [],
            classifier=classifier,
            projection_map=None,
            serializer=serializer,
        )
        == []
    )

    updates = build_tool_updates(
        [
            ModelResponse(
                parts=[ToolCallPart("final_result", {"value": "done"}, tool_call_id="final-1")],
            ),
            ModelRequest(parts=[ToolReturnPart("final_result", "done")]),
        ],
        classifier=classifier,
        projection_map=None,
        serializer=serializer,
    )

    assert updates == []



from __future__ import annotations as _annotations

from collections.abc import Iterable

__all__ = (
    "HOOKS_COMMAND_NAME",
    "MCP_SERVERS_COMMAND_NAME",
    "MODEL_COMMAND_NAME",
    "RESERVED_SLASH_COMMAND_NAMES",
    "THINKING_COMMAND_NAME",
    "TOOLS_COMMAND_NAME",
    "validate_mode_command_ids",
)

MODEL_COMMAND_NAME = "model"
THINKING_COMMAND_NAME = "thinking"
TOOLS_COMMAND_NAME = "tools"
HOOKS_COMMAND_NAME = "hooks"
MCP_SERVERS_COMMAND_NAME = "mcp-servers"
RESERVED_SLASH_COMMAND_NAMES = frozenset(
    {
        MODEL_COMMAND_NAME,
        THINKING_COMMAND_NAME,
        TOOLS_COMMAND_NAME,
        HOOKS_COMMAND_NAME,
        MCP_SERVERS_COMMAND_NAME,
    },
)


def validate_mode_command_ids(mode_ids: Iterable[str]) -> None:
    normalized_ids: list[str] = []
    for mode_id in mode_ids:
        normalized_mode_id = mode_id.strip().lower()
        if not normalized_mode_id:
            raise ValueError("Mode ids must be non-empty and slash-command compatible.")
        if any(character.isspace() for character in normalized_mode_id):
            raise ValueError(f"Mode id {mode_id!r} is invalid. Mode ids cannot contain whitespace.")
        normalized_ids.append(normalized_mode_id)
    duplicate_ids = sorted(
        mode_id for mode_id in set(normalized_ids) if normalized_ids.count(mode_id) > 1
    )
    if duplicate_ids:
        duplicate_text = ", ".join(duplicate_ids)
        raise ValueError(
            f"Mode ids must be unique after normalization. Duplicate ids: {duplicate_text}.",
        )
    conflicting_ids = sorted(set(normalized_ids) & RESERVED_SLASH_COMMAND_NAMES)
    if conflicting_ids:
        conflict_text = ", ".join(conflicting_ids)
        raise ValueError(
            "Mode ids cannot reuse reserved slash command names "
            f"({conflict_text}). Choose a more specific id outside the reserved keywords.",
        )



from __future__ import annotations as _annotations

import asyncio
from collections.abc import Sequence
from dataclasses import replace
from typing import TypeVar

from acp import run_agent
from acp.interfaces import Agent as AcpAgent
from pydantic_ai import Agent as PydanticAgent

from ..agent_source import (
    AgentFactory,
    AgentSource,
    FactoryAgentSource,
    StaticAgentSource,
)
from ..config import DEFAULT_AGENT_NAME, AdapterConfig
from ..hook_projection import HookProjectionMap
from ..projection import ProjectionMap
from .adapter import PydanticAcpAgent

AgentDepsT = TypeVar("AgentDepsT", contravariant=True)
OutputDataT = TypeVar("OutputDataT", covariant=True)

__all__ = ("create_acp_agent", "run_acp")


def create_acp_agent(
    agent: PydanticAgent[AgentDepsT, OutputDataT] | None = None,
    *,
    agent_factory: AgentFactory[AgentDepsT, OutputDataT] | None = None,
    agent_source: AgentSource[AgentDepsT, OutputDataT] | None = None,
    config: AdapterConfig | None = None,
    projection_maps: Sequence[ProjectionMap | HookProjectionMap] | None = None,
) -> AcpAgent:
    resolved_source = _resolve_agent_source(
        agent=agent,
        agent_factory=agent_factory,
        agent_source=agent_source,
    )
    resolved_config = _resolve_config(
        config=config,
        agent_name=agent.name if agent is not None else None,
        projection_maps=projection_maps,
    )
    adapter = PydanticAcpAgent(resolved_source, config=resolved_config)
    return adapter


def run_acp(
    agent: PydanticAgent[AgentDepsT, OutputDataT] | None = None,
    *,
    agent_factory: AgentFactory[AgentDepsT, OutputDataT] | None = None,
    agent_source: AgentSource[AgentDepsT, OutputDataT] | None = None,
    config: AdapterConfig | None = None,
    projection_maps: Sequence[ProjectionMap | HookProjectionMap] | None = None,
) -> None:
    adapter = create_acp_agent(
        agent=agent,
        agent_factory=agent_factory,
        agent_source=agent_source,
        config=config,
        projection_maps=projection_maps,
    )
    asyncio.run(run_agent(adapter))


def _resolve_agent_source(
    *,
    agent: PydanticAgent[AgentDepsT, OutputDataT] | None,
    agent_factory: AgentFactory[AgentDepsT, OutputDataT] | None,
    agent_source: AgentSource[AgentDepsT, OutputDataT] | None,
) -> AgentSource[AgentDepsT, OutputDataT]:
    provided_count = sum(provided is not None for provided in (agent, agent_factory, agent_source))
    if provided_count != 1:
        raise ValueError(
            "Exactly one of `agent`, `agent_factory`, or `agent_source` must be provided.",
        )
    if agent is not None:
        return StaticAgentSource(agent)
    if agent_factory is not None:
        return FactoryAgentSource(agent_factory)
    assert agent_source is not None
    return agent_source


def _resolve_config(
    *,
    config: AdapterConfig | None,
    agent_name: str | None,
    projection_maps: Sequence[ProjectionMap | HookProjectionMap] | None,
) -> AdapterConfig:
    resolved_config = config or AdapterConfig()
    if projection_maps is not None:
        tool_projection_maps: list[ProjectionMap] = []
        hook_projection_map = resolved_config.hook_projection_map
        for projection_map in projection_maps:
            if isinstance(projection_map, HookProjectionMap):
                hook_projection_map = projection_map
            else:
                tool_projection_maps.append(projection_map)
        resolved_config = replace(
            resolved_config,
            hook_projection_map=hook_projection_map,
            projection_maps=tuple(tool_projection_maps),
        )
    if agent_name is None:
        return resolved_config
    if resolved_config.agent_name != DEFAULT_AGENT_NAME:
        return resolved_config
    return replace(resolved_config, agent_name=agent_name)



from __future__ import annotations as _annotations

from typing import TYPE_CHECKING, Any, Literal

import httpx
from pydantic_ai.models.openai import OpenAIResponsesModelSettings
from pydantic_ai.providers.openai import OpenAIProvider

from .auth import CodexAuthConfig
from .client import create_codex_async_openai, create_codex_openai
from .model import CodexResponsesModel

if TYPE_CHECKING:
    from langchain_openai import ChatOpenAI

__all__ = ("create_codex_chat_openai", "create_codex_responses_model")


def create_codex_responses_model(
    model_name: str,
    *,
    config: CodexAuthConfig | None = None,
    http_client: httpx.AsyncClient | None = None,
    instructions: str,
    settings: OpenAIResponsesModelSettings | None = None,
) -> CodexResponsesModel:
    if instructions is None:
        raise ValueError(
            "`instructions` is required for Codex-backed Pydantic models. "
            "Pass an explicit system instruction string.",
        )
    client = create_codex_async_openai(config=config, http_client=http_client)
    model_settings: OpenAIResponsesModelSettings = {"openai_store": False}
    if settings is not None:
        model_settings.update(settings)
    model_settings["openai_store"] = False
    return CodexResponsesModel(
        model_name,
        default_instructions=instructions,
        provider=OpenAIProvider(openai_client=client),
        settings=model_settings,
    )


def create_codex_chat_openai(
    model_name: str,
    *,
    config: CodexAuthConfig | None = None,
    http_client: httpx.AsyncClient | None = None,
    instructions: str,
    sync_http_client: httpx.Client | None = None,
    include_response_headers: bool = False,
    model_kwargs: dict[str, Any] | None = None,
    output_version: Literal["v0", "responses/v1"] = "responses/v1",
    reasoning: dict[str, Any] | None = None,
    temperature: float | None = None,
    use_previous_response_id: bool = False,
) -> ChatOpenAI:
    """Build a Codex-backed LangChain chat model via `langchain-openai`.

    The returned model is pinned to the OpenAI Responses API and reuses the
    same Codex auth flow as the Pydantic helper path.
    """
    try:
        from langchain_openai import ChatOpenAI
    except ModuleNotFoundError as exc:
        raise ModuleNotFoundError(
            "Install the optional LangChain dependency first: "
            'uv add "codex-auth-helper[langchain]" or '
            'pip install "codex-auth-helper[langchain]".',
        ) from exc

    async_root_client = create_codex_async_openai(config=config, http_client=http_client)
    sync_root_client = create_codex_openai(config=config, http_client=sync_http_client)
    chat_model_kwargs = dict(model_kwargs or {})
    if instructions is None:
        raise ValueError(
            "`instructions` is required for Codex-backed LangChain models. "
            "Pass an explicit system instruction string.",
        )
    if "store" in chat_model_kwargs:
        raise ValueError(
            "Do not pass `model_kwargs['store']`; Codex-backed ChatOpenAI always forces "
            "`store=False`.",
        )
    if "instructions" in chat_model_kwargs:
        raise ValueError(
            "Pass `instructions` either through the dedicated parameter or "
            "`model_kwargs['instructions']`, not both.",
        )
    chat_model_kwargs["instructions"] = instructions
    chat_openai_kwargs: dict[str, Any] = {
        "model": model_name,
        "async_client": async_root_client.chat.completions,
        "client": sync_root_client.chat.completions,
        "include_response_headers": include_response_headers,
        "model_kwargs": chat_model_kwargs,
        "output_version": output_version,
        "reasoning": reasoning,
        "root_async_client": async_root_client,
        "root_client": sync_root_client,
        "store": False,
        "temperature": temperature,
        "use_previous_response_id": use_previous_response_id,
        "use_responses_api": True,
    }
    return ChatOpenAI(
        **chat_openai_kwargs,
    )



from __future__ import annotations as _annotations

import asyncio
import time
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from typing import Any, TypeAlias, cast

from acp.client.connection import ClientSideConnection
from acp.helpers import text_block, tool_content
from acp.interfaces import Agent, Client
from acp.schema import (
    AudioContentBlock,
    AuthenticateResponse,
    ClientCapabilities,
    CloseSessionResponse,
    EmbeddedResourceContentBlock,
    FileSystemCapabilities,
    ForkSessionResponse,
    HttpMcpServer,
    ImageContentBlock,
    InitializeResponse,
    ListSessionsResponse,
    LoadSessionResponse,
    McpServerStdio,
    PromptResponse,
    ResourceContentBlock,
    ResumeSessionResponse,
    SetSessionConfigOptionResponse,
    SetSessionModelResponse,
    SetSessionModeResponse,
    SseMcpServer,
    TextContentBlock,
    ToolCallProgress,
)

from .client import RemoteClientConnection, connect_remote_agent
from .config import TransportOptions

__all__ = ("RemoteProxyAgent", "connect_acp")

_HeaderPairs = Mapping[str, str] | Sequence[tuple[str, str]]
SessionConfigOptionValue: TypeAlias = bool | str


@dataclass(slots=True)
class _PromptLatencyState:
    started_at: float
    sequence: int
    first_update_ms: int | None = None
    update_count: int = 0


@dataclass(slots=True, frozen=True)
class _PromptLatencySnapshot:
    tool_call_id: str
    total_ms: int
    first_update_ms: int | None
    update_count: int


@dataclass(slots=True)
class _TransportLatencyTracker:
    _active: dict[str, _PromptLatencyState] = field(default_factory=dict)
    _sequence: int = 0

    def start_prompt(self, session_id: str) -> None:
        self._sequence += 1
        self._active[session_id] = _PromptLatencyState(
            started_at=time.perf_counter(),
            sequence=self._sequence,
        )

    def update_meta(self, session_id: str) -> dict[str, Any] | None:
        state = self._active.get(session_id)
        if state is None:
            return None
        state.update_count += 1
        elapsed_ms = _elapsed_ms(state.started_at)
        if state.first_update_ms is None:
            state.first_update_ms = elapsed_ms
        return {
            "acpremote": {
                "transport_latency": {
                    "elapsed_ms": elapsed_ms,
                    "first_update_ms": state.first_update_ms,
                    "update_count": state.update_count,
                },
            },
        }

    def finish_prompt(self, session_id: str) -> _PromptLatencySnapshot | None:
        state = self._active.pop(session_id, None)
        if state is None:
            return None
        return _PromptLatencySnapshot(
            tool_call_id=f"acpremote:latency:{session_id}:{state.sequence}",
            total_ms=_elapsed_ms(state.started_at),
            first_update_ms=state.first_update_ms,
            update_count=state.update_count,
        )


@dataclass(slots=True)
class _LatencyClient:
    delegate: Client
    tracker: _TransportLatencyTracker
    emit_latency_meta: bool

    async def request_permission(
        self,
        options: list[Any],
        session_id: str,
        tool_call: Any,
        **kwargs: Any,
    ) -> Any:
        return await self.delegate.request_permission(
            options,
            session_id,
            tool_call,
            **_merge_field_meta(kwargs, self._meta(session_id)),
        )

    async def session_update(self, session_id: str, update: Any, **kwargs: Any) -> None:
        await self.delegate.session_update(
            session_id,
            update,
            **_merge_field_meta(kwargs, self._meta(session_id)),
        )

    async def ext_method(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        return await self.delegate.ext_method(method, params)

    async def ext_notification(self, method: str, params: dict[str, Any]) -> None:
        await self.delegate.ext_notification(method, params)

    def _meta(self, session_id: str) -> dict[str, Any] | None:
        if not self.emit_latency_meta:
            return None
        return self.tracker.update_meta(session_id)


@dataclass(slots=True, kw_only=True)
class RemoteProxyAgent:
    url: str
    headers: _HeaderPairs | None = None
    bearer_token: str | None = None
    options: TransportOptions = field(default_factory=TransportOptions)
    _client: Client | None = field(default=None, init=False, repr=False)
    _remote: RemoteClientConnection | None = field(default=None, init=False, repr=False)
    _retired_remote: RemoteClientConnection | None = field(default=None, init=False, repr=False)
    _connect_lock: asyncio.Lock = field(default_factory=asyncio.Lock, init=False, repr=False)
    _latency_tracker: _TransportLatencyTracker = field(
        default_factory=_TransportLatencyTracker,
        init=False,
        repr=False,
    )
    _remote_cwd: str | None = field(default=None, init=False, repr=False)

    def on_connect(self, conn: Client) -> None:
        if self._client is not None and self._client is not conn and self._remote is not None:
            self._retired_remote = self._remote
            self._remote = None
            self._remote_cwd = None
        self._client = conn

    async def close(self) -> None:
        remote = self._remote
        self._remote = None
        self._remote_cwd = None
        try:
            await self._close_retired_remote()
        finally:
            if remote is not None:
                await remote.close()

    async def initialize(
        self,
        protocol_version: int,
        client_capabilities: Any | None = None,
        client_info: Any | None = None,
        **kwargs: Any,
    ) -> InitializeResponse:
        connection = await self._connection()
        return await connection.initialize(
            protocol_version=protocol_version,
            client_capabilities=_resolved_client_capabilities(
                client_capabilities,
                host_ownership=self.options.host_ownership,
            ),
            client_info=client_info,
            **kwargs,
        )

    async def new_session(
        self,
        cwd: str,
        mcp_servers: list[HttpMcpServer | SseMcpServer | McpServerStdio] | None = None,
        **kwargs: Any,
    ) -> Any:
        connection = await self._connection()
        return await connection.new_session(
            cwd=self._resolve_cwd(cwd),
            mcp_servers=mcp_servers,
            **kwargs,
        )

    async def load_session(
        self,
        cwd: str,
        session_id: str,
        mcp_servers: list[HttpMcpServer | SseMcpServer | McpServerStdio] | None = None,
        **kwargs: Any,
    ) -> LoadSessionResponse | None:
        connection = await self._connection()
        return await connection.load_session(
            cwd=self._resolve_cwd(cwd),
            session_id=session_id,
            mcp_servers=mcp_servers,
            **kwargs,
        )

    async def list_sessions(
        self,
        cursor: str | None = None,
        cwd: str | None = None,
        **kwargs: Any,
    ) -> ListSessionsResponse:
        connection = await self._connection()
        return await connection.list_sessions(
            cursor=cursor,
            cwd=self._resolve_optional_cwd(cwd),
            **kwargs,
        )

    async def set_session_mode(
        self,
        mode_id: str,
        session_id: str,
        **kwargs: Any,
    ) -> SetSessionModeResponse | None:
        connection = await self._connection()
        return await connection.set_session_mode(mode_id=mode_id, session_id=session_id, **kwargs)

    async def set_session_model(
        self,
        model_id: str,
        session_id: str,
        **kwargs: Any,
    ) -> SetSessionModelResponse | None:
        connection = await self._connection()
        return await connection.set_session_model(
            model_id=model_id,
            session_id=session_id,
            **kwargs,
        )

    async def set_config_option(
        self,
        config_id: str,
        session_id: str,
        value: SessionConfigOptionValue,
        **kwargs: Any,
    ) -> SetSessionConfigOptionResponse | None:
        connection = await self._connection()
        return await connection.set_config_option(
            config_id=config_id,
            session_id=session_id,
            value=value,
            **kwargs,
        )

    async def authenticate(
        self,
        method_id: str,
        **kwargs: Any,
    ) -> AuthenticateResponse | None:
        connection = await self._connection()
        return await connection.authenticate(method_id=method_id, **kwargs)

    async def prompt(
        self,
        prompt: list[
            TextContentBlock
            | ImageContentBlock
            | AudioContentBlock
            | ResourceContentBlock
            | EmbeddedResourceContentBlock
        ],
        session_id: str,
        message_id: str | None = None,
        **kwargs: Any,
    ) -> PromptResponse:
        self._latency_tracker.start_prompt(session_id)
        connection = await self._connection()
        response: PromptResponse
        try:
            response = await connection.prompt(
                prompt=prompt,
                session_id=session_id,
                message_id=message_id,
                **kwargs,
            )
        finally:
            snapshot = self._latency_tracker.finish_prompt(session_id)
            if snapshot is not None:
                await self._emit_latency_projection(session_id, snapshot)
        return response

    async def fork_session(
        self,
        cwd: str,
        session_id: str,
        mcp_servers: list[HttpMcpServer | SseMcpServer | McpServerStdio] | None = None,
        **kwargs: Any,
    ) -> ForkSessionResponse:
        connection = await self._connection()
        return await connection.fork_session(
            cwd=self._resolve_cwd(cwd),
            session_id=session_id,
            mcp_servers=mcp_servers,
            **kwargs,
        )

    async def resume_session(
        self,
        cwd: str,
        session_id: str,
        mcp_servers: list[HttpMcpServer | SseMcpServer | McpServerStdio] | None = None,
        **kwargs: Any,
    ) -> ResumeSessionResponse:
        connection = await self._connection()
        return await connection.resume_session(
            cwd=self._resolve_cwd(cwd),
            session_id=session_id,
            mcp_servers=mcp_servers,
            **kwargs,
        )

    async def close_session(
        self,
        session_id: str,
        **kwargs: Any,
    ) -> CloseSessionResponse | None:
        connection = await self._connection()
        return await connection.close_session(session_id=session_id, **kwargs)

    async def cancel(
        self,
        session_id: str,
        **kwargs: Any,
    ) -> None:
        connection = await self._connection()
        await connection.cancel(session_id=session_id, **kwargs)

    async def ext_method(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        connection = await self._connection()
        return await connection.ext_method(method=method, params=params)

    async def ext_notification(self, method: str, params: dict[str, Any]) -> None:
        connection = await self._connection()
        await connection.ext_notification(method=method, params=params)

    async def _connection(self) -> ClientSideConnection:
        remote = await self._remote_connection()
        return remote.connection

    async def _remote_connection(self) -> RemoteClientConnection:
        if self._remote is not None and self._retired_remote is None:
            return self._remote
        async with self._connect_lock:
            await self._close_retired_remote()
            if self._remote is not None:
                return self._remote
            client = self._client
            if client is None:
                raise RuntimeError("RemoteProxyAgent requires on_connect(...) before use.")
            remote_client: Client = client
            if self.options.emit_latency_meta:
                remote_client = cast(
                    "Client",
                    _LatencyClient(
                        delegate=client,
                        tracker=self._latency_tracker,
                        emit_latency_meta=True,
                    ),
                )
            remote = await connect_remote_agent(
                remote_client,
                self.url,
                options=self.options,
                headers=self.headers,
                bearer_token=self.bearer_token,
            )
            self._remote_cwd = remote.metadata.remote_cwd if remote.metadata is not None else None
            self._remote = remote
            return remote

    async def _close_retired_remote(self) -> None:
        remote = self._retired_remote
        self._retired_remote = None
        if remote is not None:
            await remote.close()

    def _resolve_cwd(self, cwd: str) -> str:
        return self._remote_cwd or cwd

    def _resolve_optional_cwd(self, cwd: str | None) -> str | None:
        if self._remote_cwd is not None:
            return self._remote_cwd
        return cwd

    async def _emit_latency_projection(
        self,
        session_id: str,
        snapshot: _PromptLatencySnapshot,
    ) -> None:
        if not self.options.emit_latency_projection:
            return
        client = self._client
        if client is None:
            return
        await client.session_update(
            session_id,
            ToolCallProgress(
                tool_call_id=snapshot.tool_call_id,
                session_update="tool_call_update",
                title="Transport Latency",
                kind="other",
                status="completed",
                content=[tool_content(text_block(_format_latency_summary(snapshot)))],
                raw_output={
                    "acpremote": {
                        "transport_latency": {
                            "total_ms": snapshot.total_ms,
                            "first_update_ms": snapshot.first_update_ms,
                            "update_count": snapshot.update_count,
                        },
                    },
                },
            ),
            source="acpremote-latency",
        )


def connect_acp(
    url: str,
    *,
    bearer_token: str | None = None,
    headers: _HeaderPairs | None = None,
    options: TransportOptions | None = None,
) -> Agent:
    return cast(
        "Agent",
        RemoteProxyAgent(
            url=url,
            bearer_token=bearer_token,
            headers=headers,
            options=options or TransportOptions(),
        ),
    )


def _elapsed_ms(started_at: float) -> int:
    return max(0, int((time.perf_counter() - started_at) * 1000))


def _merge_field_meta(
    kwargs: dict[str, Any],
    extra: dict[str, Any] | None,
) -> dict[str, Any]:
    if extra is None:
        return kwargs
    field_meta = kwargs.get("field_meta")
    if field_meta is None:
        return {**kwargs, "field_meta": extra}
    if not isinstance(field_meta, dict):
        return kwargs
    merged = dict(field_meta)
    for key, value in extra.items():
        current = merged.get(key)
        if isinstance(current, dict) and isinstance(value, dict):
            merged[key] = {**current, **value}
        else:
            merged[key] = value
    return {**kwargs, "field_meta": merged}


def _format_latency_summary(snapshot: _PromptLatencySnapshot) -> str:
    first_update_text = (
        f"{snapshot.first_update_ms} ms"
        if snapshot.first_update_ms is not None
        else "no streamed updates"
    )
    return (
        f"First remote update: {first_update_text}\n"
        f"Turn complete: {snapshot.total_ms} ms\n"
        f"Streamed updates: {snapshot.update_count}"
    )


def _resolved_client_capabilities(
    client_capabilities: Any | None,
    *,
    host_ownership: str,
) -> Any | None:
    if host_ownership == "client_passthrough":
        return client_capabilities
    if not isinstance(client_capabilities, ClientCapabilities):
        return client_capabilities
    fs = client_capabilities.fs
    sanitized_fs = None
    if fs is not None:
        sanitized_fs = FileSystemCapabilities(
            field_meta=fs.field_meta,
            read_text_file=False,
            write_text_file=False,
        )
    return ClientCapabilities(
        field_meta=client_capabilities.field_meta,
        auth=client_capabilities.auth,
        fs=sanitized_fs,
        terminal=False,
    )



from __future__ import annotations as _annotations

import importlib
import runpy
import sys
from itertools import cycle
from pathlib import Path
from types import SimpleNamespace
from typing import Any, cast

import pytest
from langchain_acp import AcpSessionContext
from langchain_acp.session import utc_now
from langchain_core.language_models import GenericFakeChatModel
from langchain_core.messages import AIMessage


def _fake_codex_model() -> GenericFakeChatModel:
    return GenericFakeChatModel(messages=cycle([AIMessage(content="codex-ready")]))


def _load_example_module(
    monkeypatch: pytest.MonkeyPatch,
    module_name: str,
) -> Any:
    import codex_auth_helper

    monkeypatch.setattr(
        codex_auth_helper,
        "create_codex_chat_openai",
        lambda _model_name, **_kwargs: _fake_codex_model(),
    )
    sys.modules.pop(module_name, None)
    return importlib.import_module(module_name)


def _run_example_module_as_main(module_name: str) -> dict[str, Any]:
    sys.modules.pop(module_name, None)
    return runpy.run_module(module_name, run_name="__main__")


def test_langchain_example_main_dispatches_run_acp(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    workspace_graph = _load_example_module(monkeypatch, "examples.langchain.workspace_graph")
    captured: list[tuple[Any, Any]] = []

    def fake_run_acp(*, graph_factory: Any, config: Any) -> None:
        captured.append((graph_factory, config))

    monkeypatch.setattr(workspace_graph, "run_acp", fake_run_acp)
    workspace_graph.main()

    assert captured == [(workspace_graph.graph_from_session, workspace_graph.config)]


def test_codex_langchain_example_builds_graph_from_helper(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    codex_graph = _load_example_module(monkeypatch, "examples.langchain.codex_graph")
    captured: dict[str, Any] = {}

    def fake_create_codex_chat_openai(model_name: str, *, instructions: str) -> str:
        captured["model_name"] = model_name
        captured["instructions"] = instructions
        return "codex-model"

    def fake_create_agent(*, model: Any, tools: list[Any], name: str) -> object:
        captured["model"] = model
        captured["tools"] = tools
        captured["name"] = name
        return object()

    monkeypatch.setattr(codex_graph, "create_codex_chat_openai", fake_create_codex_chat_openai)
    monkeypatch.setattr(codex_graph, "create_agent", fake_create_agent)

    graph = codex_graph.build_graph()

    assert graph is not None
    assert captured["model_name"] == codex_graph.MODEL_NAME
    assert captured["model"] == "codex-model"
    assert captured["name"] == "codex-graph"
    assert "workspace assistant" in captured["instructions"]
    assert [tool.__name__ for tool in captured["tools"]] == ["describe_codex_surface"]
    assert "Codex graph features:" in codex_graph.describe_codex_surface()
    assert codex_graph.config.available_models
    assert codex_graph.config.available_modes


def test_codex_langchain_example_main_dispatches_run_acp(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    codex_graph = _load_example_module(monkeypatch, "examples.langchain.codex_graph")
    captured: list[tuple[Any, Any]] = []

    def fake_run_acp(*, graph_factory: Any, config: Any) -> None:
        captured.append((graph_factory, config))

    monkeypatch.setattr(codex_graph, "run_acp", fake_run_acp)

    codex_graph.main()

    assert captured == [(codex_graph.graph_from_session, codex_graph.config)]


def test_codex_langchain_example_module_runs_as_main(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    observed: dict[str, Any] = {}

    import codex_auth_helper
    import langchain.agents
    import langchain_acp

    monkeypatch.setattr(
        codex_auth_helper,
        "create_codex_chat_openai",
        lambda _, **_kwargs: "codex-model",
    )
    monkeypatch.setattr(
        langchain.agents,
        "create_agent",
        lambda *, model, tools, name: {
            "model": model,
            "tools": tools,
            "name": name,
        },
    )

    def fake_run_acp(*, graph_factory: Any, config: Any) -> None:
        observed["call"] = (graph_factory, config)

    monkeypatch.setattr(langchain_acp, "run_acp", fake_run_acp)

    _run_example_module_as_main("examples.langchain.codex_graph")

    graph_factory, config = observed["call"]
    session = AcpSessionContext(
        session_id="codex-example",
        cwd=Path.cwd(),
        created_at=utc_now(),
        updated_at=utc_now(),
    )
    graph = graph_factory(session)
    assert graph["model"] == "codex-model"
    assert graph["name"].startswith("codex-")
    assert config is not None


def test_codex_langchain_example_graph_factory_uses_session_model_and_mode(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    codex_graph = _load_example_module(monkeypatch, "examples.langchain.codex_graph")
    captured: dict[str, Any] = {}

    def fake_create_codex_chat_openai(model_name: str, *, instructions: str) -> str:
        captured["model_name"] = model_name
        captured["instructions"] = instructions
        return "codex-model"

    def fake_create_agent(*, model: Any, tools: list[Any], name: str) -> object:
        captured["model"] = model
        captured["name"] = name
        captured["tools"] = tools
        return {"model": model, "name": name}

    monkeypatch.setattr(codex_graph, "create_codex_chat_openai", fake_create_codex_chat_openai)
    monkeypatch.setattr(codex_graph, "create_agent", fake_create_agent)

    session = AcpSessionContext(
        session_id="codex-session",
        cwd=tmp_path,
        created_at=utc_now(),
        updated_at=utc_now(),
        session_model_id="gpt-5.4",
        session_mode_id="plan",
    )
    graph = codex_graph.graph_from_session(session)

    assert graph == {"model": "codex-model", "name": f"codex-plan-{tmp_path.name}"}
    assert captured["model_name"] == "gpt-5.4"
    assert "short plan" in captured["instructions"]


def test_codex_langchain_example_edit_mode_instructions(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    codex_graph = _load_example_module(monkeypatch, "examples.langchain.codex_graph")

    assert "changed files" in codex_graph.codex_instructions(mode_id="edit")


def test_langchain_example_workspace_helpers_cover_seeded_paths(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    workspace_graph = _load_example_module(monkeypatch, "examples.langchain.workspace_graph")
    root = tmp_path / "agent_demos" / "workspace-graph"
    monkeypatch.setattr(workspace_graph, "WORKSPACE_ROOT", root)

    workspace_graph._ensure_workspace()

    assert workspace_graph.list_workspace_files() == "README.md"
    assert "session-aware `graph_from_session(...)`" in workspace_graph.describe_workspace_surface()
    assert "Workspace Graph Demo" in workspace_graph.read_workspace_note("README.md")
    assert workspace_graph.write_workspace_note("scratch.txt", "# Hello") == "Wrote `scratch.txt`."
    assert workspace_graph.list_workspace_files() == "README.md\nscratch.txt"
    assert workspace_graph.config.available_models
    assert workspace_graph.config.available_modes

    with pytest.raises(ValueError, match="workspace graph demo directory"):
        workspace_graph._resolve_workspace_path("../escape.md")

    with pytest.raises(ValueError, match="File not found"):
        workspace_graph.read_workspace_note("missing.md")

    assert workspace_graph.config.projection_maps


def test_langchain_example_workspace_graph_factory_uses_session_root(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    workspace_graph = _load_example_module(monkeypatch, "examples.langchain.workspace_graph")
    session_root = tmp_path / "remote-workspace"
    session_root.mkdir(parents=True, exist_ok=True)
    session = AcpSessionContext(
        session_id="workspace-example",
        cwd=session_root,
        created_at=utc_now(),
        updated_at=utc_now(),
    )

    graph = workspace_graph.graph_from_session(session)

    assert graph is not None
    seeded_root = session_root / "agent_demos" / "workspace-graph"
    assert (seeded_root / "README.md").exists()


def test_langchain_example_workspace_graph_factory_uses_session_model_and_mode(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    workspace_graph = _load_example_module(monkeypatch, "examples.langchain.workspace_graph")
    captured: dict[str, Any] = {}

    def fake_create_codex_chat_openai(model_name: str, *, instructions: str) -> str:
        captured["model_name"] = model_name
        captured["instructions"] = instructions
        return "workspace-model"

    def fake_create_agent(*, model: Any, tools: list[Any], name: str) -> object:
        captured["model"] = model
        captured["name"] = name
        captured["tools"] = tools
        return {"model": model, "name": name}

    monkeypatch.setattr(workspace_graph, "create_codex_chat_openai", fake_create_codex_chat_openai)
    monkeypatch.setattr(workspace_graph, "create_agent", fake_create_agent)

    session = AcpSessionContext(
        session_id="workspace-session",
        cwd=tmp_path,
        created_at=utc_now(),
        updated_at=utc_now(),
        session_model_id="gpt-5.4",
        session_mode_id="edit",
    )
    graph = workspace_graph.graph_from_session(session)

    assert graph == {
        "model": "workspace-model",
        "name": f"workspace-edit-{tmp_path.name}",
    }
    assert captured["model_name"] == "gpt-5.4"
    assert "smallest viable file update" in captured["instructions"]


def test_langchain_example_workspace_plan_mode_instructions(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    workspace_graph = _load_example_module(monkeypatch, "examples.langchain.workspace_graph")

    assert "short implementation plan" in workspace_graph.codex_instructions(mode_id="plan")


def test_langchain_example_workspace_bound_tools_cover_private_closures(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    workspace_graph = _load_example_module(monkeypatch, "examples.langchain.workspace_graph")
    root = tmp_path / "agent_demos" / "workspace-graph"
    root.mkdir(parents=True, exist_ok=True)
    tools = {tool.__name__: tool for tool in workspace_graph._bind_workspace_tools(root)}

    assert tools["describe_workspace_surface"]() == workspace_graph.describe_workspace_surface()
    assert tools["list_workspace_files"]() == ""
    assert tools["read_workspace_note"].__name__ == "read_workspace_note"
    assert tools["write_workspace_note"].__name__ == "write_workspace_note"

    assert tools["write_workspace_note"]("note.md", "# Demo") == "Wrote `note.md`."
    assert tools["list_workspace_files"]() == "README.md\nnote.md"
    assert tools["read_workspace_note"]("note.md") == "# Demo"

    with pytest.raises(ValueError, match="File not found"):
        tools["read_workspace_note"]("missing.md")


def test_langchain_example_workspace_module_runs_as_main(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    observed: dict[str, Any] = {}

    import langchain_acp

    def fake_run_acp(*, graph_factory: Any, config: Any) -> None:
        observed["call"] = (graph_factory, config)

    monkeypatch.setattr(langchain_acp, "run_acp", fake_run_acp)
    import codex_auth_helper

    monkeypatch.setattr(
        codex_auth_helper,
        "create_codex_chat_openai",
        lambda _model_name, **_kwargs: _fake_codex_model(),
    )

    _run_example_module_as_main("examples.langchain.workspace_graph")

    graph_factory, config = observed["call"]
    assert graph_factory.__name__ == "graph_from_session"
    assert config is not None


def test_deepagents_example_main_dispatches_run_acp(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    deepagents_graph = _load_example_module(monkeypatch, "examples.langchain.deepagents_graph")
    captured: list[tuple[Any, Any]] = []

    def fake_run_acp(*, graph_factory: Any, config: Any) -> None:
        captured.append((graph_factory, config))

    monkeypatch.setattr(deepagents_graph, "run_acp", fake_run_acp)
    deepagents_graph.main()

    assert captured == [(deepagents_graph.graph_from_session, deepagents_graph.config)]


def test_deepagents_example_workspace_helpers_cover_seeded_paths(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    deepagents_graph = _load_example_module(monkeypatch, "examples.langchain.deepagents_graph")
    root = tmp_path / "agent_demos" / "deepagents-graph"
    monkeypatch.setattr(deepagents_graph, "WORKSPACE_ROOT", root)

    deepagents_graph._ensure_workspace()

    assert deepagents_graph.list_workspace_files() == "brief.md\nnotes.md"
    assert "DeepAgents Demo" in deepagents_graph.read_file("brief.md")
    assert (
        deepagents_graph.write_file("itinerary.md", "# Trip")
        == "Mock write accepted for itinerary.md"
    )
    assert deepagents_graph.list_workspace_files() == "brief.md\nnotes.md"
    assert deepagents_graph.config.available_models
    assert deepagents_graph.config.available_modes
    assert (
        deepagents_graph._resolve_workspace_path("itinerary.md", root=root)
        == (root / "itinerary.md").resolve()
    )
    assert (
        deepagents_graph._resolve_workspace_path("deepagents-graph/itinerary.md", root=root)
        == (root / "itinerary.md").resolve()
    )
    assert (
        deepagents_graph._resolve_workspace_path(str((root / "itinerary.md").resolve()), root=root)
        == (root / "itinerary.md").resolve()
    )
    assert (
        deepagents_graph._resolve_workspace_path(
            str((root.parent / "itinerary.md").resolve()),
            root=root,
        )
        == (root / "itinerary.md").resolve()
    )

    with pytest.raises(ValueError, match="DeepAgents example workspace"):
        deepagents_graph._resolve_workspace_path("../escape.md")

    with pytest.raises(ValueError, match="File not found"):
        deepagents_graph.read_file("missing.md")

    assert deepagents_graph.config.projection_maps


def test_deepagents_example_helper_edges_cover_remaining_branches(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    deepagents_graph = _load_example_module(monkeypatch, "examples.langchain.deepagents_graph")
    root = tmp_path / "agent_demos" / "deepagents-graph"
    monkeypatch.setattr(deepagents_graph, "WORKSPACE_ROOT", root)

    assert "direct file edits" in deepagents_graph.codex_instructions(mode_id="edit")

    outside_root = root.parent / "outside"
    outside_root.mkdir(parents=True, exist_ok=True)
    outside_child = outside_root / "escape.md"
    resolved = deepagents_graph._resolve_workspace_path(str(outside_child.resolve()), root=root)
    assert resolved == (root / "outside" / "escape.md").resolve()

    nested_prefixed = root.parent / root.name / "nested" / "note.md"
    normalized = deepagents_graph._resolve_workspace_path(str(nested_prefixed.resolve()), root=root)
    assert normalized == (root / "nested" / "note.md").resolve()

    external_absolute = Path("/tmp") / "deepagents-escape.md"
    with pytest.raises(ValueError, match="DeepAgents example workspace"):
        deepagents_graph._resolve_workspace_path(str(external_absolute.resolve()), root=root)

    with pytest.raises(ValueError, match="mocked workspace file"):
        deepagents_graph._normalized_mock_path("")


def test_deepagents_example_graph_factory_requires_optional_dependency(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    deepagents_graph = _load_example_module(monkeypatch, "examples.langchain.deepagents_graph")
    session = AcpSessionContext(
        session_id="example-session",
        cwd=Path.cwd(),
        created_at=utc_now(),
        updated_at=utc_now(),
    )

    monkeypatch.setattr(deepagents_graph, "_deepagents_available", lambda: False)

    with pytest.raises(RuntimeError, match="langchain-acp\\[deepagents\\]"):
        deepagents_graph.graph_from_session(session)


def test_deepagents_example_graph_factory_builds_graph_from_lazy_import(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    deepagents_graph = _load_example_module(monkeypatch, "examples.langchain.deepagents_graph")
    root = tmp_path / "agent_demos" / "deepagents-graph"
    monkeypatch.setattr(deepagents_graph, "WORKSPACE_ROOT", root)
    monkeypatch.setattr(deepagents_graph, "_deepagents_available", lambda: True)

    captured: dict[str, Any] = {}

    def fake_create_deep_agent(**kwargs: Any) -> object:
        captured.update(kwargs)
        return object()

    monkeypatch.setitem(
        sys.modules,
        "deepagents",
        cast("Any", SimpleNamespace(create_deep_agent=fake_create_deep_agent)),
    )

    session = AcpSessionContext(
        session_id="example-session",
        cwd=root,
        created_at=utc_now(),
        updated_at=utc_now(),
    )
    graph = deepagents_graph.graph_from_session(session)

    assert graph is not None
    assert captured["interrupt_on"] == {"write_file": True}
    assert captured["name"] == "deepagents-ask-deepagents-graph"
    tool_names = {tool.__name__ for tool in cast("list[Any]", captured["tools"])}
    assert {
        "list_workspace_files",
        "read_file",
        "write_file",
        "acp_get_plan",
        "acp_set_plan",
        "acp_update_plan_entry",
        "acp_mark_plan_done",
    } <= tool_names


def test_deepagents_example_graph_factory_builds_graph_when_dependency_is_mocked(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    deepagents_graph = _load_example_module(monkeypatch, "examples.langchain.deepagents_graph")
    fake_graph = object()
    captured: dict[str, Any] = {}

    def fake_create_deep_agent(**kwargs: Any) -> object:
        captured.update(kwargs)
        return fake_graph

    monkeypatch.setattr(deepagents_graph, "_deepagents_available", lambda: True)
    monkeypatch.setitem(
        sys.modules,
        "deepagents",
        cast("Any", SimpleNamespace(create_deep_agent=fake_create_deep_agent)),
    )

    session = AcpSessionContext(
        session_id="example-session",
        cwd=tmp_path,
        created_at=utc_now(),
        updated_at=utc_now(),
        session_model_id="gpt-5.4",
        session_mode_id="plan",
    )

    assert deepagents_graph.graph_from_session(session) is fake_graph
    tool_names = {tool.__name__ for tool in cast("list[Any]", captured["tools"])}
    assert {
        "list_workspace_files",
        "read_file",
        "write_file",
        "acp_get_plan",
        "acp_set_plan",
        "acp_update_plan_entry",
        "acp_mark_plan_done",
    } <= tool_names
    assert captured["name"] == f"deepagents-plan-{tmp_path.name}"


def test_deepagents_example_bound_tools_use_session_workspace(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    deepagents_graph = _load_example_module(monkeypatch, "examples.langchain.deepagents_graph")
    session_root = tmp_path / "remote-workspace"
    session_root.mkdir(parents=True, exist_ok=True)
    bound_root = session_root / "agent_demos" / "deepagents-graph"
    tools = {tool.__name__: tool for tool in deepagents_graph._bind_workspace_tools(bound_root)}

    assert tools["list_workspace_files"]() == "brief.md\nnotes.md"
    assert tools["write_file"]("notes.md", "# Deep") == "Mock write accepted for notes.md"
    assert tools["list_workspace_files"]() == "brief.md\nnotes.md"
    assert "Mock Notes" in tools["read_file"]("notes.md")

    with pytest.raises(ValueError, match="File not found"):
        tools["read_file"]("missing.md")


def test_deepagents_example_seed_session_and_module_run_as_main(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    deepagents_graph = _load_example_module(monkeypatch, "examples.langchain.deepagents_graph")
    root = tmp_path / "agent_demos" / "deepagents-graph"
    monkeypatch.setattr(deepagents_graph, "WORKSPACE_ROOT", root)

    session = deepagents_graph._seed_session()
    assert session.session_id == "deepagents-example"
    assert session.cwd == root.resolve()

    observed: dict[str, Any] = {}

    import langchain_acp

    def fake_run_acp(*, graph_factory: Any, config: Any) -> None:
        observed["call"] = (graph_factory, config)

    monkeypatch.setattr(langchain_acp, "run_acp", fake_run_acp)
    import codex_auth_helper

    monkeypatch.setattr(
        codex_auth_helper,
        "create_codex_chat_openai",
        lambda _model_name, **_kwargs: _fake_codex_model(),
    )
    _run_example_module_as_main("examples.langchain.deepagents_graph")

    graph_factory, config = observed["call"]
    assert graph_factory.__name__ == "graph_from_session"
    assert config is not None



from __future__ import annotations as _annotations

import argparse
import asyncio
import importlib
import os
import sys
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Any, TypeGuard

from acp import run_agent
from acp.interfaces import Agent

from .command import CommandOptions
from .config import build_server_paths
from .proxy_agent import connect_acp
from .server import serve_acp, serve_stdio_command

__all__ = ("AcpRemoteCliError", "main")


class AcpRemoteCliError(RuntimeError):
    """Raised for user-facing CLI failures."""


@dataclass(frozen=True, slots=True)
class _TargetRef:
    module_name: str
    attribute_path: str | None


def main(argv: Sequence[str] | None = None) -> int:
    parser = _build_parser()
    try:
        args = parser.parse_args(list(argv) if argv is not None else None)
        args.handler(args)
    except AcpRemoteCliError as exc:
        print(f"Error: {exc}", file=sys.stderr)
        return 2
    except KeyboardInterrupt:
        return 130
    return 0


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="acpremote",
        description="Expose or mirror ACP agents over WebSocket transport.",
    )
    subcommands = parser.add_subparsers(dest="command_name", required=True)

    serve_parser = subcommands.add_parser(
        "serve",
        help="Expose a native ACP Python target over WebSocket.",
    )
    serve_parser.add_argument("target", help="Python target, for example `my_app:agent`.")
    _add_server_options(serve_parser)
    serve_parser.add_argument(
        "-p",
        "--path",
        dest="import_roots",
        action="append",
        default=[],
        help="Extra import root for loading the target module. Can be repeated.",
    )
    serve_parser.set_defaults(handler=_handle_serve)

    expose_parser = subcommands.add_parser(
        "expose",
        help="Expose an ACP-over-stdio command over WebSocket.",
    )
    _add_server_options(expose_parser)
    expose_parser.add_argument("--cwd", help="Working directory for the child command.")
    expose_parser.add_argument(
        "--env",
        dest="env_overrides",
        action="append",
        default=[],
        metavar="KEY=VALUE",
        help="Environment override for the child command. Can be repeated.",
    )
    expose_parser.add_argument(
        "--stderr-mode",
        choices=("inherit", "discard"),
        default="inherit",
        help="How to handle child process stderr.",
    )
    expose_parser.add_argument(
        "--terminate-timeout",
        type=float,
        default=5.0,
        help="Seconds to wait before killing a closing child process.",
    )
    expose_parser.add_argument(
        "command",
        nargs=argparse.REMAINDER,
        help="Command to run. Use `--` before commands that have their own flags.",
    )
    expose_parser.set_defaults(handler=_handle_expose)

    mirror_parser = subcommands.add_parser(
        "mirror",
        help="Mirror a remote ACP WebSocket endpoint back to local stdio ACP.",
    )
    mirror_parser.add_argument("addr", help="Remote ACP WebSocket address.")
    _add_token_options(mirror_parser)
    mirror_parser.set_defaults(handler=_handle_mirror)

    return parser


def _add_server_options(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--host",
        default="127.0.0.1",
        help="Host interface for the WebSocket server.",
    )
    parser.add_argument(
        "--port",
        default=8080,
        type=int,
        help="TCP port for the WebSocket server.",
    )
    parser.add_argument(
        "--mount-path",
        default="/acp",
        help="Mount path for metadata and WebSocket routes.",
    )
    _add_token_options(parser)


def _add_token_options(parser: argparse.ArgumentParser) -> None:
    token_group = parser.add_mutually_exclusive_group()
    token_group.add_argument("--bearer-token", help="Bearer token value.")
    token_group.add_argument(
        "--token-env",
        help="Environment variable containing the bearer token.",
    )


def _handle_serve(args: argparse.Namespace) -> None:
    agent = _load_native_agent_target(
        str(args.target),
        import_roots=tuple(str(path) for path in args.import_roots),
    )
    bearer_token = _resolve_bearer_token(args)

    async def run() -> None:
        server = await serve_acp(
            agent,
            host=str(args.host),
            port=int(args.port),
            mount_path=str(args.mount_path),
            bearer_token=bearer_token,
        )
        await _serve_forever(server, host=str(args.host), mount_path=str(args.mount_path))

    asyncio.run(run())


def _handle_expose(args: argparse.Namespace) -> None:
    command = _normalize_command(tuple(str(part) for part in args.command))
    bearer_token = _resolve_bearer_token(args)
    env = _parse_env_overrides(tuple(str(value) for value in args.env_overrides))
    command_options = CommandOptions(
        command=command,
        cwd=str(args.cwd) if args.cwd is not None else None,
        env=env,
        stderr_mode=args.stderr_mode,
        terminate_timeout=float(args.terminate_timeout),
    )

    async def run() -> None:
        server = await serve_stdio_command(
            command_options,
            host=str(args.host),
            port=int(args.port),
            mount_path=str(args.mount_path),
            bearer_token=bearer_token,
        )
        await _serve_forever(server, host=str(args.host), mount_path=str(args.mount_path))

    asyncio.run(run())


def _handle_mirror(args: argparse.Namespace) -> None:
    agent = connect_acp(str(args.addr), bearer_token=_resolve_bearer_token(args))
    asyncio.run(run_agent(agent))


async def _serve_forever(server: Any, *, host: str, mount_path: str) -> None:
    _print_server_banner(server, host=host, mount_path=mount_path)
    try:
        await server.serve_forever()
    finally:
        server.close()
        await server.wait_closed()


def _print_server_banner(server: Any, *, host: str, mount_path: str) -> None:
    port = _server_port(server)
    paths = build_server_paths(mount_path)
    if port is None:
        print(
            f"Serving ACP WebSocket at ws://{host}{paths.websocket_path}",
            file=sys.stderr,
        )
        return
    print(
        f"Serving ACP WebSocket at ws://{host}:{port}{paths.websocket_path}",
        file=sys.stderr,
    )


def _server_port(server: Any) -> int | None:
    sockets = getattr(server, "sockets", None)
    if not sockets:
        return None
    socket = sockets[0]
    address = socket.getsockname()
    if isinstance(address, tuple) and len(address) >= 2 and isinstance(address[1], int):
        return address[1]
    return None


def _resolve_bearer_token(args: argparse.Namespace) -> str | None:
    bearer_token = getattr(args, "bearer_token", None)
    if bearer_token:
        return str(bearer_token)
    token_env = getattr(args, "token_env", None)
    if token_env is None:
        return None
    value = os.getenv(str(token_env), "").strip()
    if value:
        return value
    raise AcpRemoteCliError(f"`{token_env}` is not set or is empty.")


def _normalize_command(command: tuple[str, ...]) -> tuple[str, ...]:
    normalized = command[1:] if command[:1] == ("--",) else command
    if not normalized:
        raise AcpRemoteCliError("Command must not be empty.")
    return normalized


def _parse_env_overrides(values: tuple[str, ...]) -> dict[str, str] | None:
    if not values:
        return None
    env: dict[str, str] = {}
    for value in values:
        key, separator, env_value = value.partition("=")
        if not key or separator == "":
            raise AcpRemoteCliError("Environment overrides must use `KEY=VALUE`.")
        env[key] = env_value
    return env


def _load_native_agent_target(target: str, *, import_roots: Sequence[str]) -> Agent:
    loaded_target = _load_target(target, import_roots=import_roots)
    if _is_acp_agent(loaded_target):
        return loaded_target
    raise AcpRemoteCliError(
        "Target must resolve to a native `acp.interfaces.Agent`. "
        "Use `acpkit serve` for Pydantic or LangChain targets.",
    )


def _load_target(target: str, *, import_roots: Sequence[str]) -> Any:
    reference = _parse_target_ref(target)
    module = _import_target_module(reference, target=target, import_roots=import_roots)
    if reference.attribute_path is None:
        return _resolve_latest_native_agent(module, target)
    return _resolve_attribute(module, reference.attribute_path, target=target)


def _parse_target_ref(target: str) -> _TargetRef:
    module_name, separator, attribute_path = target.partition(":")
    if not module_name:
        raise AcpRemoteCliError(
            "Target must include a module name, for example `my_app` or `my_app:agent`.",
        )
    if separator and not attribute_path:
        raise AcpRemoteCliError("Target attribute cannot be empty.")
    return _TargetRef(
        module_name=module_name,
        attribute_path=attribute_path or None,
    )


def _import_target_module(
    reference: _TargetRef,
    *,
    target: str,
    import_roots: Sequence[str],
) -> ModuleType:
    _ensure_import_root(Path.cwd())
    for import_root in import_roots:
        _ensure_import_root(Path(import_root))
    importlib.invalidate_caches()
    try:
        return importlib.import_module(reference.module_name)
    except ImportError as exc:
        raise AcpRemoteCliError(
            f"Could not import module `{reference.module_name}` from target `{target}`.",
        ) from exc


def _resolve_latest_native_agent(module: ModuleType, target: str) -> Agent:
    latest_target: Agent | None = None
    for value in vars(module).values():
        if _is_acp_agent(value):
            latest_target = value
    if latest_target is None:
        raise AcpRemoteCliError(
            f"Target `{target}` did not resolve to a native ACP agent and the module defines no "
            "native ACP agent instance.",
        )
    return latest_target


def _resolve_attribute(module: ModuleType, attribute_path: str, *, target: str) -> Any:
    value: Any = module
    for attribute_name in attribute_path.split("."):
        try:
            value = getattr(value, attribute_name)
        except AttributeError as exc:
            raise AcpRemoteCliError(
                f"Target `{target}` is missing attribute `{attribute_name}`.",
            ) from exc
    return value


def _ensure_import_root(path: Path) -> None:
    root_path = path.parent if path.exists() and path.is_file() else path
    resolved_path = str(root_path.resolve())
    if resolved_path not in sys.path:
        sys.path.insert(0, resolved_path)


def _is_acp_agent(value: Any) -> TypeGuard[Agent]:
    required_methods = (
        "initialize",
        "new_session",
        "prompt",
        "cancel",
        "on_connect",
    )
    return all(callable(getattr(value, method_name, None)) for method_name in required_methods)
